import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_circular import bend_circular
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight
from gdsfactory.cross_section import strip
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def coupler90(
    gap: float = 0.2,
    radius: float = 10.0,
    bend: ComponentFactory = bend_euler,
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> Component:
    r"""straight coupled to a bend.

    Args:
        gap: um
        radius: um
        straight: for straight
        bend: for bend
        cross_section:
        kwargs: cross_section settings

    .. code::

             3
             |
            /
           /
        2_/
        1____4

    """
    c = Component()
    x = cross_section(radius=radius, **kwargs)

    bend90 = (
        bend(cross_section=cross_section, radius=radius, **kwargs)
        if callable(bend)
        else bend
    )
    bend_ref = c << bend90
    straight_component = (
        straight(
            cross_section=cross_section,
            length=bend90.ports["o2"].midpoint[0] - bend90.ports["o1"].midpoint[0],
            **kwargs
        )
        if callable(straight)
        else straight
    )

    wg_ref = c << straight_component
    width = x.info["width"]

    pbw = bend_ref.ports["o1"]
    bend_ref.movey(pbw.midpoint[1] + gap + width)

    c.absorb(wg_ref)
    c.absorb(bend_ref)

    c.add_port("o1", port=wg_ref.ports["o1"])
    c.add_port("o4", port=wg_ref.ports["o2"])
    c.add_port("o2", port=bend_ref.ports["o1"])
    c.add_port("o3", port=bend_ref.ports["o2"])
    return c


def coupler90circular(bend: ComponentFactory = bend_circular, **kwargs):
    return coupler90(bend=bend, **kwargs)


if __name__ == "__main__":
    # c = coupler90circular(gap=0.3)
    # c << coupler90(gap=0.3)
    c = coupler90(radius=3, layer=(2, 0))
    c.show()
    c.pprint()
    # print(c.ports)



""" You can define both DOEs and placer information in a YAML file
All the placer information need to be nested inside a placer section

```yaml
iso_lines_coarse1:
  component: ISO_COARS_OPT1
  settings:
    dx: [50.]

  placer:
    type: pack_row/ pack_col/ grid / fixed_coords
        pack_grid:
    x0: 0
    y0: 0
    align_x: W
    align_y: S
    next_to: iso_lines_coarse1
```

"""
import collections
import os
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import klayout.db as pya
import numpy as np
from klayout.dbcore import Cell, CellInstArray, Layout
from numpy import float64
from omegaconf import OmegaConf

from gdsfactory.autoplacer import text
from gdsfactory.autoplacer.helpers import CELLS, import_cell, load_gds
from gdsfactory.config import CONFIG

UM_TO_GRID = 1e3
DEFAULT_BBOX_LAYER_IGNORE = [(8484, 8484)]


def _print(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def to_grid(x: float64, um_to_grid: int = UM_TO_GRID) -> int:
    return int(x * um_to_grid)


class SizeInfo:
    def __init__(
        self,
        cell: Cell,
        layout: Optional[Layout] = None,
        ignore_layers: List[Tuple[int, int]] = DEFAULT_BBOX_LAYER_IGNORE,
        um_to_grid: int = UM_TO_GRID,
    ) -> None:
        """
        layout is required if cell is a cell reference instead of a cell
        """
        if isinstance(cell, pya.CellInstArray):
            bbox = cell.bbox(layout)
        else:
            parent_layout = cell.layout()
            layer_indexes = parent_layout.layer_indexes()

            for process_id, purpose_id in ignore_layers:
                layer_id = parent_layout.find_layer(process_id, purpose_id)
                if layer_id in layer_indexes:
                    layer_indexes.remove(layer_id)

            bbox = None
            for layer_id in layer_indexes:
                per_layer_bbox = cell.bbox_per_layer(layer_id)
                if bbox is None:
                    bbox = per_layer_bbox
                else:
                    bbox = bbox + per_layer_bbox

        self.box = bbox

        self.west = self.box.left / um_to_grid
        self.east = self.box.right / um_to_grid

        self.south = self.box.bottom / um_to_grid
        self.north = self.box.top / um_to_grid

        self.width = self.east - self.west
        self.height = self.north - self.south

        xc = int(0.5 * (self.east + self.west))
        yc = int(0.5 * (self.north + self.south))

        self.sw = np.array([self.west, self.south])
        self.se = np.array([self.east, self.south])
        self.nw = np.array([self.west, self.north])
        self.ne = np.array([self.east, self.north])

        self.cw = np.array([self.west, yc])
        self.ce = np.array([self.east, yc])
        self.nc = np.array([xc, self.north])
        self.sc = np.array([xc, self.south])
        self.cc = self.center = np.array([xc, yc])

    @property
    def rect(self):
        w, e, s, n = self.west, self.east, self.south, self.north
        return [(w, s), (e, s), (e, n), (w, n)]

    def __str__(self):
        return "w: {}\ne: {}\ns: {}\nn: {}\n".format(
            self.west, self.east, self.south, self.north
        )


def placer_grid_cell_refs(
    cells,
    cols=1,
    rows=1,
    dx=10.0,
    dy=10.0,
    x0=0,
    y0=0,
    um_to_grid=UM_TO_GRID,
    **settings,
):
    """cells: list of cells - order matters for placing"""

    indices = [(i, j) for j in range(cols) for i in range(rows)]

    if rows * cols < len(cells):
        raise ValueError(
            "Shape ({}, {}): Not enough emplacements ({}) for all these components"
            " ({}).".format(rows, cols, len(indices), len(cells))
        )
    components = []
    for cell, (i, j) in zip(cells, indices):
        _x = int((x0 + j * dx) * um_to_grid)
        _y = int((y0 + i * dy) * um_to_grid)

        transform = pya.Trans(_x, _y)
        c_ref = pya.CellInstArray(cell.cell_index(), transform)
        components += [c_ref]

    return components


def pack_row(
    cells: List[Cell],
    row_ids: Optional[List[int]] = None,
    nb_cols: Optional[int] = None,
    x0: Union[float, int] = 0,
    y0: Union[float, int] = 0,
    align_x: str = "W",
    align_y: str = "S",
    margin: Union[float, int] = 20,
    margin_x: Optional[Union[float, int]] = None,
    margin_y: Optional[Union[float, int]] = None,
    um_to_grid: int = UM_TO_GRID,
    period_x: Optional[float] = None,
    period_y: Optional[float] = None,
    rotation: int = 0,
) -> List[CellInstArray]:
    """Pack row.
    Args:
        cells: a list of cells (size n)
        row_ids: a list of row ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells in the same row
        nb_cols: number of columns
        period_x, period_y: not used by default,
            if set, use this period instead of computing the component spacing
            from the margin and the component dimension

    Returns:list of cell references
    """
    si_list = [SizeInfo(c, um_to_grid=um_to_grid) for c in cells]
    heights = [si.height for si in si_list]
    margin_y = margin_y if margin_y is not None else margin
    margin_x = margin_x if margin_x is not None else margin

    if row_ids is None:
        row_ids = []
        nb_cells = len(cells)
        if nb_cols is None:
            nb_cols = len(cells)
        nb_full_rows = nb_cells // nb_cols
        nb_cols_last_row = nb_cells % nb_cols
        for row_id in range(nb_full_rows):
            row_ids += [row_id] * nb_cols

        last_row_index = row_id + 1
        row_ids += [last_row_index] * nb_cols_last_row

    if len(cells) != len(row_ids):
        raise ValueError(
            "Each cell should be assigned a row id.         Got {} cells for {} row ids".format(
                len(cells), len(row_ids)
            )
        )

    # Find the height of each row to fit the cells
    # Also group the cells by row

    unique_row_ids = list(set(row_ids))
    unique_row_ids.sort()
    _row_to_heights = {r: [] for r in set(row_ids)}
    row_to_cells = {r: [] for r in unique_row_ids}
    for row, h, cell in zip(row_ids, heights, cells):
        _row_to_heights[row] += [h]
        row_to_cells[row] += [cell]

    row_to_height = {k: max(v) for k, v in _row_to_heights.items()}

    components = []

    # Do the packing per row
    y = y0
    for row in unique_row_ids:
        cells = row_to_cells[row]
        x = x0

        for c in cells:
            si = SizeInfo(c, um_to_grid=um_to_grid)
            if align_x == "W" and align_y == "S":
                component_origin = si.sw
            elif align_x == "E" and align_y == "S":
                component_origin = si.se
            elif align_x == "E" and align_y == "N":
                component_origin = si.ne
            elif align_x == "W" and align_y == "N":
                component_origin = si.nw
            try:
                _x = to_grid(x - component_origin[0], um_to_grid)
                _y = to_grid(y - component_origin[1], um_to_grid)

                transform = pya.Trans(rotation / 2, 0, _x, _y)
                # transform = pya.Trans(_x, _y)
                c_ref = pya.CellInstArray(c.cell_index(), transform)
                components += [c_ref]

            except BaseException:
                print(x, component_origin[0], um_to_grid)
                print("ISSUE PLACING AT", _x, _y)
                if align_x not in ["W", "E"]:
                    _print("align_x should be `W`, `E` or a float")
                if align_y not in ["N", "S"]:
                    _print("align_y should be `N`, `S` or a float")
                # raise

            dx = si.width + margin_x if period_x is None else period_x
            if align_x == "W":
                x += dx
            else:
                x -= dx

        dy = row_to_height[row] + margin_y if period_y is None else period_y

        if align_y == "S":
            y += dy
        else:
            y -= dy

    return components


def pack_col(
    cells: List[Cell],
    col_ids: None = None,
    nb_rows: Optional[int] = None,
    x0: float = 0,
    y0: float = 0,
    align_x: str = "W",
    align_y: str = "S",
    margin: int = 20,
    margin_x: Optional[int] = None,
    margin_y: Optional[int] = None,
    um_to_grid: int = UM_TO_GRID,
    period_x: Optional[float] = None,
    period_y: Optional[float] = None,
) -> List[CellInstArray]:
    """

    Args:
        cells: a list of cells  (size n)
        col_ids: a list of column ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells are packed in the same column

    Returns:
        list of cell references
    """
    widths = [SizeInfo(c, um_to_grid=um_to_grid).width for c in cells]
    margin_y = margin_y if margin_y is not None else margin
    margin_x = margin_x if margin_x is not None else margin

    if col_ids is None:
        col_ids = []
        nb_cells = len(cells)
        if nb_rows is None:
            nb_rows = len(cells)
        nb_full_cols = nb_cells // nb_rows
        nb_rows_last_col = nb_cells % nb_rows
        for col_id in range(nb_full_cols):
            col_ids += [col_id] * nb_rows

        last_col_index = col_id + 1
        col_ids += [last_col_index] * nb_rows_last_col

    if len(cells) != len(col_ids):
        raise ValueError(
            "Each cell should be assigned a row id.         Got {} cells for {} col ids".format(
                len(cells), len(col_ids)
            )
        )

    # Find the width of each column to fit the cells
    # Also group the cells by column

    unique_col_ids = list(set(col_ids))
    unique_col_ids.sort()
    _col_to_widths = {r: [] for r in set(col_ids)}
    col_to_cells = {r: [] for r in unique_col_ids}
    for col, w, cell in zip(col_ids, widths, cells):
        _col_to_widths[col] += [w]
        col_to_cells[col] += [cell]

    col_to_width = {k: max(v) for k, v in _col_to_widths.items()}

    components = []

    # Do the packing per column
    x = x0
    for col in unique_col_ids:
        cells = col_to_cells[col]
        y = y0

        for c in cells:
            si = SizeInfo(c, um_to_grid=um_to_grid)
            if align_x == "W" and align_y == "S":
                component_origin = si.sw
            elif align_x == "E" and align_y == "S":
                component_origin = si.se
            elif align_x == "E" and align_y == "N":
                component_origin = si.ne
            elif align_x == "W" and align_y == "N":
                component_origin = si.nw

            _x = to_grid(x - component_origin[0], um_to_grid=um_to_grid)
            _y = to_grid(y - component_origin[1], um_to_grid=um_to_grid)

            try:
                transform = pya.Trans(_x, _y)
                c_ref = pya.CellInstArray(c.cell_index(), transform)
                components += [c_ref]
            except BaseException:
                print(x, component_origin[0], um_to_grid)
                print(y, component_origin[1], um_to_grid)
                print("ISSUE PLACING AT", _x, _y)
                print("ISSUE PLACING at", _x, _y)

            dy = si.height + margin_y if period_y is None else period_y
            if align_y == "S":
                y += dy
            else:
                y -= dy

        dx = col_to_width[col] + margin_x if period_x is None else period_x
        if align_x == "W":
            x += dx
        else:
            x -= dx

    return components


def placer_fixed_coords(
    cells, x, y, x0=0, y0=0, do_permutation=False, um_to_grid=UM_TO_GRID, **kwargs
):
    """place cells using a list of coordinates"""

    # List all coordinates
    if do_permutation:
        coords = [(_x, _y) for _x in x for _y in y]
    else:
        coords = [(_x, _y) for _x, _y in zip(x, y)]

    # Update origin
    coords = [(c[0] + x0, c[1] + y0) for c in coords]

    # Generate cell list
    if len(cells) == 1:
        cells = cells * len(coords)

    # update coordinates from um to grid
    coords = [(int(c[0] * um_to_grid), int(c[1] * um_to_grid)) for c in coords]

    # Generate transforms
    transforms = [pya.Trans(*c) for c in coords]

    return [pya.CellInstArray(c.cell_index(), t) for c, t in zip(cells, transforms)]


def load_yaml(filepath: Path) -> Any:
    """load placer settings

    Args:
        filepath: a yaml file containing the does and placer information

    Returns:
        a dictionnary of DOEs with:
        {
            doe_name1: {...}
            doe_name2: {...}
            ...
        }

    """

    does = {}
    data = OmegaConf.load(filepath)
    data = OmegaConf.to_container(data)
    mask = data.pop("mask")

    if "layer_doe_label" not in mask:
        mask["layer_doe_label"] = (102, 6)

    for doe_name, doe in data.items():
        _doe = {}
        _doe.update(doe)
        does[doe_name] = _doe
    return does, mask


DOE_CELLS = {}


def load_doe(doe_name: str, doe_root: Path) -> List[Layout]:
    """
    Load all components for this DOE from the cache
    """
    doe_dir = os.path.join(doe_root, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")

    if os.path.isfile(content_file):
        with open(content_file) as f:
            lines = f.read().split("\n")
            line = lines[0]

            if line.startswith("TEMPLATE:"):
                """
                If using a template, load the GDS from DOE folder used as a template
                """
                template_name = line.split(":")[1].strip()
                return load_doe(template_name, doe_root)

            else:
                """
                Otherwise load the GDS from the current folder
                """
                component_names = line.split(" , ")
                gdspaths = [
                    os.path.join(doe_dir, name + ".gds") for name in component_names
                ]
                cells = [load_gds(gdspath) for gdspath in gdspaths]

        # print("LOAD DOE")
        # for _c in cells:
        # print(_c.top_cell().name)
        # print()

        return cells


PLACER_NAME2FUNC = {
    "grid": placer_grid_cell_refs,
    "pack_row": pack_row,
    "pack_col": pack_col,
    "fixed_coords": placer_fixed_coords,
}


def separate_does_from_templates(dicts: Dict[str, Any]) -> Any:
    templates = {}
    does = {}
    for name, d in dicts.items():
        if "type" in d.keys():  # and d["type"] == "template":
            templates[name] = d
        else:
            does[name] = d

    # We do not want to propagate "type": template to the does => removing it here
    for d in templates.values():
        d.pop("type")

    return does, templates


def update_dicts_recurse(
    target_dict: Dict[str, Any], default_dict: Dict[str, Any]
) -> Dict[str, Any]:
    target_dict = target_dict.copy()
    default_dict = default_dict.copy()
    for k, v in default_dict.items():
        if k not in target_dict:
            vtype = type(v)
            if vtype == dict or vtype == collections.OrderedDict:
                target_dict[k] = v.copy()  # To avoid issues when popping
            else:
                target_dict[k] = v
        else:
            vtype = type(target_dict[k])
            if vtype == dict or vtype == collections.OrderedDict:
                target_dict[k] = update_dicts_recurse(target_dict[k], default_dict[k])
    return target_dict


def place_from_yaml(
    filepath_yaml: Path,
    root_does: Path = CONFIG["cache_doe_directory"],
    precision: float = 1e-9,
    fontpath: Path = text.FONT_PATH,
    default_align_x: str = "W",
    default_align_y: str = "S",
    default_margin: int = 10,
    default_x0: str = "E",
    default_y0: str = "S",
) -> Cell:
    """Returns a gds cell composed of DOEs/components given in a yaml file
    allows for each DOE to have its own x and y spacing (more flexible than method1)

    Args:
        filepath_yaml:
        root_does: used for cache, requires content.txt
    """
    transform_identity = pya.Trans(0, 0)
    dicts, mask_settings = load_yaml(filepath_yaml)

    does, templates = separate_does_from_templates(dicts)

    placed_doe = None
    placed_does = {}
    top_level_name = mask_settings.get("name", "TOP_LEVEL")
    layer_doe_label = mask_settings["layer_doe_label"]
    top_level_layout = pya.Layout()

    # Set database units according to precision
    top_level_layout.dbu = precision / 1e-6
    dbu = top_level_layout.dbu
    um_to_grid = int(1 / dbu)

    top_level = top_level_layout.create_cell(top_level_name)
    global CELLS
    CELLS[top_level_name] = top_level_layout

    default_doe_settings = {
        "add_doe_label": False,
        "add_doe_visual_label": False,
        "dx_visual_label": 0,
        "dy_visual_label": 0,
    }

    for doe_name, doe in does.items():

        # If a template is specified, apply it
        if "template" in doe:
            doe_templates = doe["template"]
            if type(doe_templates) != list:
                doe_templates = [doe_templates]
            for doe_template in doe_templates:
                try:
                    doe = update_dicts_recurse(doe, templates[doe_template])

                except BaseException:
                    print(doe_template, "does not exist")
                    raise
        doe = update_dicts_recurse(doe, default_doe_settings)

        # Get all the components
        components = load_doe(doe_name, root_does)

        # Check that the high level components are all unique
        # For now this is mostly to circumvent a bug
        # But the design manual also specifies that DOE components should have
        # unique names. So one instance per cell

        if components:
            if len(components) != len(set([_c.top_cell().name for _c in components])):
                __dict_component_debug = {}
                for _c in components:
                    _name = _c.top_cell().name
                    if _name not in __dict_component_debug:
                        __dict_component_debug[_name] = 0
                    __dict_component_debug[_name] += 1
                duplicates_components = [
                    _name
                    for _name, _count in __dict_component_debug.items()
                    if _count > 1
                ]
                print("Please remove duplicate components at DOE entry level: ")
                print(duplicates_components)

            components = [
                import_cell(top_level_layout, _c.top_cell()) for _c in components
            ]

        default_placer_settings = {
            "align_x": default_align_x,
            "align_y": default_align_y,
            "margin": default_margin,
            "x0": default_x0,
            "y0": default_y0,
        }
        settings = default_placer_settings.copy()
        placer = doe.get("placer")

        if placer:
            placer_type = placer.pop("type", "pack_col")
            settings.update(doe["placer"])
        else:
            placer_type = "pack_col"

        if placer_type not in PLACER_NAME2FUNC:
            raise ValueError(
                f"{placer_type} is not an available placer, Choose:"
                f" {list(PLACER_NAME2FUNC.keys())}"
            )
        _placer = PLACER_NAME2FUNC[placer_type]

        # All other attributes are assumed to be settings for the placer

        # Check if the cell should be attached to a specific parent cell
        if "parent" in settings:
            parent_name = settings.pop("parent")
            if parent_name not in CELLS:
                # Create parent cell in layout and insert it under top level
                parent_cell = top_level_layout.create_cell(parent_name)
                CELLS[parent_name] = parent_cell
                parent_cell_instance = pya.CellInstArray(
                    parent_cell.cell_index(), transform_identity
                )
                top_level.insert(parent_cell_instance)
            doe_parent_cell = CELLS[parent_name]
        else:
            # If no parent specified, insert the DOE at top level
            doe_parent_cell = top_level

        # Check if we should create a DOE cell which regroups the DOEs
        if "with_doe_cell" in settings:
            with_doe_cell = settings.pop("with_doe_cell")
        else:
            with_doe_cell = True

        # x0, y0 can either be float or string
        x0 = settings.pop("x0")
        y0 = settings.pop("y0")

        # Check whether we are doing relative or absolute placement
        # if (x0 in ["E", "W"] or y0 in ["N", "S"]) and not placed_doe:
        #     raise ValueError(
        #         "At least one DOE must be placed to use relative placement"
        #     )

        # For relative placement (to previous DOE)
        if "margin_x" not in settings:
            settings["margin_x"] = settings["margin"]
        if "margin_y" not in settings:
            settings["margin_y"] = settings["margin"]

        if "inter_margin_x" not in settings:
            inter_margin_x = settings["margin_x"]
        else:
            inter_margin_x = settings.pop("inter_margin_x")

        if "inter_margin_y" not in settings:
            inter_margin_y = settings["margin_y"]
        else:
            inter_margin_y = settings.pop("inter_margin_y")

        align_x = settings["align_x"]
        align_y = settings["align_y"]

        # Make sure that the alignment is sensible depending on how we stack

        # If we specify a DOE to place next to, use it
        if "next_to" in settings:
            placed_doe = placed_does[settings.pop("next_to")]

        # print(placed_doe)
        # print(placed_does)

        # Otherwise, use previously placed DOE as starting point
        doe_si = (
            SizeInfo(placed_doe, top_level_layout, um_to_grid=um_to_grid)
            if placed_doe is not None
            else None
        )
        if x0 == "E":
            x0 = doe_si.east
            if align_x == "W":
                x0 += inter_margin_x

        if x0 == "W":
            x0 = doe_si.west
            if align_x == "E":
                x0 -= inter_margin_x

        if y0 == "N":
            y0 = doe_si.north
            if align_y == "S":
                y0 += inter_margin_y

        if y0 == "S":
            y0 = doe_si.south
            if align_y == "N":
                y0 -= inter_margin_y

        # Add x0, y0 in settings as float
        settings["x0"] = x0
        settings["y0"] = y0

        settings["um_to_grid"] = um_to_grid

        placed_components = _placer(components, **settings)

        # Place components within a cell having the DOE name

        if with_doe_cell or len(placed_components) > 1:
            doe_cell = top_level_layout.create_cell(doe_name)
            CELLS[doe_name] = doe_cell
            for instance in placed_components:
                doe_cell.insert(instance)
            placed_does[doe_name] = doe_cell
            placed_doe = doe_cell
            doe_instance = pya.CellInstArray(doe_cell.cell_index(), transform_identity)
        else:
            # If only single cell and we want to skip the sweep cell
            doe_instance = placed_components[0]
            placed_does[doe_name] = doe_instance
            placed_doe = doe_instance

        add_doe_label = doe["add_doe_label"]
        add_doe_visual_label = doe["add_doe_visual_label"]

        if add_doe_label:
            label_layer_index, label_layer_datatype = layer_doe_label
            layer_index = top_level.layout().insert_layer(
                pya.LayerInfo(label_layer_index, label_layer_datatype)
            )
            # Add the name of the DOE at the center of the cell
            _p = doe_instance.bbox(top_level_layout).center()
            _text = pya.Text(doe_name, _p.x, _p.y)
            top_level.shapes(layer_index).insert(_text)

        if add_doe_visual_label:
            _bbox = doe_instance.bbox(top_level_layout)

            idbu = 1 / top_level.layout().dbu
            x_text = _bbox.center().x + doe["dx_visual_label"] * idbu
            y_text = _bbox.bottom + (15.0 + doe["dy_visual_label"]) * idbu
            _text = text.add_text(
                top_level, doe_name, position=(x_text, y_text), fontpath=fontpath
            )
            # _transform = pya.DTrans(x_text, y_text)
            # top_level.insert(pya.CellInstArray(_text.cell_index(), _transform))

        doe_parent_cell.insert(doe_instance)

    return top_level


def place_and_write(
    filepath_yaml, root_does=CONFIG["cache_doe_directory"], filepath_gds="top_level.gds"
):
    c = place_from_yaml(filepath_yaml, root_does)
    _print("writing...")
    c.write(filepath_gds)


def assemble_subdies_from_yaml(filepath, subdies_directory, mask_directory=None):
    data = OmegaConf.load(filepath)
    data = OmegaConf.to_container(data)

    mask = data.pop("mask")
    mask_name = mask["name"]

    # Remaining entries are subdies
    dict_subdies = {
        k: (v["x"], v["y"], v["R"] if "R" in v else 0) for k, v in data.items()
    }

    return assemble_subdies(mask_name, dict_subdies, subdies_directory, mask_directory)


def assemble_subdies(
    mask_name,
    dict_subdies,
    subdies_directory,
    mask_directory=None,
    um_to_grid=UM_TO_GRID,
):
    """
    Args:
        dict_subdies: {subdie_name: (x, y, rotation) in (um, um, deg)}
        subdies_directory: directory where the subdies should be looked for
    """
    top_level_layout = pya.Layout()
    top_level = top_level_layout.create_cell(mask_name)
    if mask_directory is None:
        mask_directory = subdies_directory

    for subdie_name, (x_um, y_um, R) in dict_subdies.items():
        gdspath = os.path.join(subdies_directory, subdie_name + ".gds")
        subdie = load_gds(gdspath).top_cell()

        _subdie = import_cell(top_level_layout, subdie)

        t = pya.Trans(R / 2, 0, int(x_um * um_to_grid), int(y_um * um_to_grid))
        # t = pya.Trans(0, 0)
        subdie_instance = pya.CellInstArray(_subdie.cell_index(), t)
        top_level.insert(subdie_instance)

    top_level.write(os.path.join(mask_directory, mask_name + ".gds"))
    return top_level


def _demo():
    import gdsfactory as gf

    c = gf.components.straight()
    gdspath = c.write_gds_with_metadata()

    layout1 = load_gds(gdspath)
    cell1 = layout1.top_cell()
    cell1_instance1 = pya.CellInstArray(cell1.cell_index(), pya.Trans(10, 0))

    layout2 = pya.Layout()
    layout2.create_cell("TOP_LEVEL")

    layout2.cell("TOP_LEVEL").insert(cell1_instance1)
    layout2.write("test.gds")


if __name__ == "__main__":
    _demo()
    print(CELLS)



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.routing.get_bundle import get_bundle_same_axis_no_grouping


def test_link_optical_ports_no_grouping(
    data_regression: DataRegressionFixture, check: bool = True
) -> Component:

    c = gf.Component("test_link_optical_ports_no_grouping")
    w = c << gf.components.straight_array(n=4, spacing=200)
    d = c << gf.components.nxn(west=4, east=1)
    d.y = w.y
    d.xmin = w.xmax + 200

    ports1 = [
        w.ports["o7"],
        w.ports["o8"],
    ]
    ports2 = [
        d.ports["o2"],
        d.ports["o1"],
    ]

    routes = get_bundle_same_axis_no_grouping(ports1, ports2, sort_ports=True)
    # routes = gf.routing.get_bundle(ports1, ports2, sort_ports=True)

    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
    return c


if __name__ == "__main__":
    c = test_link_optical_ports_no_grouping(None, check=False)
    c.show()



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler_ring import coupler_ring
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper
from gdsfactory.config import call_if_func
from gdsfactory.snap import assert_on_2nm_grid

taper2 = gf.partial(taper, width2=3)


@gf.cell
def ring_single_dut(
    component=taper2,
    wg_width=0.5,
    gap=0.2,
    length_x=4,
    radius=5,
    length_y=0,
    coupler=coupler_ring,
    straight=straight_function,
    bend=bend_euler,
    with_component=True,
    **kwargs
):
    """Single bus ring made of two couplers (ct: top, cb: bottom)
    connected with two vertical straights (wyl: left, wyr: right)
    (Device Under Test) in the middle to extract loss from quality factor


    Args:
        with_component: if False changes component for just a straight

    .. code::

          bl-wt-br
          |      | length_y
          wl     component
          |      |
         --==cb==-- gap

          length_x
    """
    component = call_if_func(component)
    assert_on_2nm_grid(gap)

    coupler = call_if_func(coupler, gap=gap, radius=radius, length_x=length_x, **kwargs)
    straight_side = call_if_func(
        straight, width=wg_width, length=length_y + component.xsize, **kwargs
    )
    straight_top = call_if_func(straight, width=wg_width, length=length_x, **kwargs)
    bend = call_if_func(bend, width=wg_width, radius=radius, **kwargs)

    c = Component()
    c.component = component
    cb = c << coupler
    wl = c << straight_side
    if with_component:
        d = c << component
    else:
        d = c << straight_side
    bl = c << bend
    br = c << bend
    wt = c << straight_top

    wl.connect(port="o2", destination=cb.ports["o2"])
    bl.connect(port="o2", destination=wl.ports["o1"])

    wt.connect(port="o1", destination=bl.ports["o1"])
    br.connect(port="o2", destination=wt.ports["o2"])
    d.connect(port="o1", destination=br.ports["o1"])

    c.add_port("o2", port=cb.ports["o4"])
    c.add_port("o1", port=cb.ports["o1"])
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    c = ring_single_dut()
    c.show()



"""equations from paper https://doi.org/10.1364/OE.27.010456
"""

dn_dt_si = 1.87e-4
dn_dt_sio2 = 8.5e-6


def delta_temperature_pi(length: float, wavelength: float = 1.55, dndT=1.8e-4):
    return wavelength / (2 * length * dndT)


if __name__ == "__main__":
    for length in [320, 600]:
        dT = delta_temperature_pi(length=length)
        print(f"length = {length}, dT = {dT:.3f} K")



"""add_pin adss a Pin to a port, add_pins adds Pins to all ports:

- pins
- outline

Some functions modify a component without changing its name.
Make sure these functions are inside a new Component or called as a decorator
They without modifying the cell name

"""
import json
from typing import Callable, Optional, Tuple

import numpy as np
from numpy import ndarray
from omegaconf import OmegaConf

import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.component import Component, ComponentReference
from gdsfactory.port import Port
from gdsfactory.tech import LAYER


def _rotate(v: ndarray, m: ndarray) -> ndarray:
    return np.dot(m, v)


def add_pin_triangle(
    component: Component,
    port: Port,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.TEXT,
) -> None:
    """Add triangle pin with a right angle, pointing out of the port

    Args:
        component:
        port: Port
        layer: for the pin marker
        label_layer: for the label
    """
    p = port

    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dtip = np.array([d, 0])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptip = p.position + _rotate(dtip, rot_mat)
    polygon = [p0, p1, ptip]
    component.add_polygon(polygon, layer=layer)

    if label_layer:
        component.add_label(
            text=str(p.name),
            position=ptip,
            layer=label_layer,
        )


def add_pin_square_inside(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.TEXT,
) -> None:
    """Add square pin towards the inside of the port

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label

    .. code::

           _______________
          |               |
          |               |
          |               |
          ||              |
          ||              |
          |               |
          |      __       |
          |_______________|


    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([-pin_length, -d])
    dtopin = np.array([-pin_length, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)
    if label_layer:
        component.add_label(
            text=str(p.name),
            position=p.midpoint,
            layer=label_layer,
        )


def add_pin_square_double(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.TEXT,
) -> None:
    """Add two square pins: one inside with label, one outside

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label

    .. code::

           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    # outer square
    d = p.width / 2
    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([pin_length / 2, -d])
    dtopin = np.array([pin_length / 2, +d])
    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    # inner square
    d = p.width / 2
    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([-pin_length / 2, -d])
    dtopin = np.array([-pin_length / 2, +d])
    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    x = (p0[0] + ptopin[0]) / 2
    y = (ptopin[1] + pbotin[1]) / 2
    if label_layer:
        component.add_label(
            text=str(p.name),
            position=(x, y),
            layer=label_layer,
        )


def add_pin_square(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.PORT,
    port_margin: float = 0.0,
) -> None:
    """Add half out pin to a component.

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label
        port_margin: margin to port edge


    .. code::

           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2 + port_margin

    dbot = np.array([pin_length / 2, -d])
    dtop = np.array([pin_length / 2, d])
    dbotin = np.array([-pin_length / 2, -d])
    dtopin = np.array([-pin_length / 2, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    if label_layer:
        component.add_label(
            text=str(p.name),
            position=p.midpoint,
            layer=label_layer,
        )


def add_outline(
    component: Component,
    reference: Optional[ComponentReference] = None,
    layer: Tuple[int, int] = LAYER.DEVREC,
    **kwargs,
) -> None:
    """Adds devices outline bounding box in layer.

    Args:
        component: where to add the markers
        reference: to read outline from
        layer: to add padding
        default: default padding
        top: North padding
        bottom
        right
        left
    """
    c = reference or component
    if hasattr(component, "parent"):
        component = component.parent
    points = get_padding_points(component=c, default=0, **kwargs)
    component.add_polygon(points, layer=layer)


def add_pins(
    component: Component,
    reference: Optional[ComponentReference] = None,
    function: Callable = add_pin_square_inside,
    select_ports: Optional[Callable] = None,
    **kwargs,
) -> None:
    """Add Pin port markers.

    Be careful with this function as it modifies the component

    Args:
        component: to add ports to
        reference:
        function: to add each pin
        select_ports: function to select_ports
        kwargs: add pins function settings

    """
    reference = reference or component
    ports = (
        select_ports(reference.ports).values()
        if select_ports
        else reference.get_ports_list()
    )
    for port in ports:
        function(component=component, port=port, **kwargs)


add_pins_triangle = gf.partial(add_pins, function=add_pin_triangle)


@gf.cell
def add_pins_container(
    component: Component,
    add_pins_function=add_pins_triangle,
) -> Component:
    c = Component()
    ref = c << component
    add_pins_function(component=c, reference=ref)
    c.ref = ref
    return c


def add_settings_label(
    component: Component,
    reference: ComponentReference,
    label_layer: Tuple[int, int] = LAYER.LABEL_SETTINGS,
) -> None:
    """Add settings in label

    Args:
        componnent
        reference
        label_layer:

    """
    settings_dict = OmegaConf.to_container(reference.settings.full)
    settings_string = f"settings={json.dumps(settings_dict)}"
    print(settings_string)
    if len(settings_string) > 1024:
        raise ValueError(f"label > 1024 characters: {settings_string}")
    component.add_label(
        position=reference.center, text=settings_string, layer=label_layer
    )


def add_instance_label(
    component: Component,
    reference: ComponentReference,
    instance_name: Optional[str] = None,
    layer: Tuple[int, int] = LAYER.LABEL_INSTANCE,
) -> None:
    """Adds label to a reference in a component."""

    instance_name = (
        instance_name
        or f"{reference.parent.name},{int(reference.x)},{int(reference.y)}"
    )
    x = gf.snap.snap_to_grid(reference.x)
    y = gf.snap.snap_to_grid(reference.y)

    component.add_label(
        text=instance_name,
        position=(x, y),
        layer=layer,
    )


def add_pins_and_outline(
    component: Component,
    reference: ComponentReference,
    add_outline_function: Optional[Callable] = add_outline,
    add_pins_function: Optional[Callable] = add_pins,
    add_settings_function: Optional[Callable] = add_settings_label,
    add_instance_label_function: Optional[Callable] = add_settings_label,
) -> None:
    """Add markers:
    - outline
    - pins for the ports
    - label for the name
    - label for the settings

    Args:
        component: where to add the markers
        reference
        add_outline_function
        add_pins_function: to add pins to ports
        add_settings_function: to add outline around the component
        add_instance_label_function: labels each instance

    """
    if add_outline_function:
        add_outline_function(component=component, reference=reference)
    if add_pins_function:
        add_pins_function(component=component, reference=reference)
    if add_settings_function:
        add_settings_function(component=component, reference=reference)
    if add_instance_label_function:
        add_instance_label_function(component=component, reference=reference)


@gf.cell
def add_pins_to_references(
    component: Component,
    function: Callable = add_pins_triangle,
) -> Component:
    """Add pins to Component references.

    Args:
        component: component to add pins to
        references:
        function: function to add pins
    """
    c = Component()
    c.add_ref(component)

    references = component.references

    for reference in references:
        function(component=c, reference=reference)

    return c


def test_add_pins() -> None:
    c1 = gf.c.straight(length=2)
    c2 = add_pins_container(component=c1)

    p1 = len(c1.get_polygons())
    p2 = len(c2.get_polygons())

    assert p2 == p1 + 2
    assert len(c2.polygons) == 2


if __name__ == "__main__":
    # c = test_add_pins()
    # c.show()

    # c = gf.c.straight(length=2)
    # c = add_pins_container(component=c)

    # c.show(show_ports_suborts=True)

    # c2 = add_pins_container(component=c1)

    # p1 = len(c1.get_polygons())
    # p2 = len(c2.get_polygons())
    # assert p2 == p1 + 2

    # c1 = gf.c.straight_heater_metal(length=2)
    # c2 = add_pins_to_references(c1)
    c = gf.c.ring_single()
    # cc = add_pins_to_references(component=c)
    # cc.show(show_ports=False)
    c.show(show_subports=True)
    c.show(show_ports=True)



from typing import Optional

import numpy as np
import picwriter.components as pc

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.waveguide_template import strip
from gdsfactory.types import Coordinate, Coordinates, Floats, Layer


@cell
def grating_coupler_circular(
    taper_angle: float = 30.0,
    taper_length: float = 10.0,
    length: float = 30.0,
    period: float = 1.0,
    fill_factor: float = 0.7,
    port: Coordinate = (0.0, 0.0),
    layer: Layer = gf.LAYER.WG,
    layer_slab: Optional[Layer] = None,
    layer_cladding: Layer = gf.LAYER.WGCLAD,
    teeth_list: Optional[Coordinates] = None,
    direction: str = "EAST",
    polarization: str = "te",
    wavelength: float = 1.55,
    fiber_marker_width: float = 11.0,
    fiber_marker_layer: Optional[Layer] = gf.LAYER.TE,
    wg_width: float = 0.5,
    cladding_offset: float = 2.0,
) -> Component:
    r"""Returns Grating coupler from Picwriter

    Args:
        taper_angle: taper flare angle in degrees
        taper_length: Length of the taper before the grating coupler.
        length: total grating coupler length.
        period: Grating period.
        fill_factor: (period-gap)/period.
        port: Cartesian coordinate of the input port
        layer: waveguide layer
        layer_slab: slab layer for partial etched gratings
        layer_cladding: for the cladding (using cladding_offset)
        teeth_list: (gap, width) tuples to be used as the gap and teeth widths
          for irregularly spaced gratings.
          For example, [(0.6, 0.2), (0.7, 0.3), ...] would be a gap of 0.6,
          then a tooth of width 0.2, then gap of 0.7 and tooth of 0.3, and so on.
          Overrides *period*, *dutycycle*, and *length*.  Defaults to None.
        direction: Direction that the component will point *towards*,
          can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`,
          OR an angle (float, in radians)
        polarization: te or tm
        wavelength: wavelength um
        fiber_marker_width:
        wg_width: for input waveguide
        cladding_offset:


    .. code::

        side view
                      fiber
                   /  /  /  /
                  /  /  /  /
                _|-|_|-|_|-|___
        WG  o1  ______________|

                      /
        top view     / |
                    /| |
                   / | |
                  /taper_angle
                 /_ _| |
        wg_width |   | |
                 \   | |
                  \  | |
                   \ | |
                    \| |
                 <-->
                taper_length
    """

    c = pc.GratingCoupler(
        gf.call_if_func(
            strip,
            cladding_offset=cladding_offset,
            wg_width=wg_width,
            layer=layer,
            layer_cladding=layer_cladding,
        ),
        theta=np.deg2rad(taper_angle),
        length=length,
        taper_length=taper_length,
        period=period,
        dutycycle=fill_factor,
        ridge=True if layer_slab else False,
        ridge_layers=layer_slab,
        teeth_list=teeth_list,
        port=port,
        direction=direction,
    )

    c = gf.read.from_picwriter(c)
    c.info.polarization = polarization
    c.info.wavelength = wavelength
    x = np.round(c.center[0] + taper_length / 2, 3)

    if fiber_marker_layer:
        circle = gf.components.circle(
            radius=fiber_marker_width / 2, layer=fiber_marker_layer
        )
        circle_ref = c.add_ref(circle)
        circle_ref.movex(x)

    c.add_port(
        name=f"vertical_{polarization.lower()}",
        midpoint=[x, 0],
        width=fiber_marker_width,
        orientation=0,
        layer=fiber_marker_layer,
    )

    c.auto_rename_ports()
    return c


_gap_width = tuple([0.1] * 10 + [0.5] * 10)


@cell
def grating_coupler_circular_arbitrary(teeth_list: Floats = _gap_width, **kwargs):
    """Returns grating coupler,
    teeth list is on a single list that starts with gap, width, gap ...

    Args:
        teeth_list: list of gaps and widths
        kwargs: for grating_coupler_circular
    """
    teeth_list = zip(teeth_list[::2], teeth_list[1::2])
    return grating_coupler_circular(teeth_list=list(teeth_list), **kwargs)


if __name__ == "__main__":
    # c = grating_coupler_circular_arbitrary(taper_length=30, layers_slab=((2,0), (3,0)))
    # c = grating_coupler_circular_arbitrary(taper_length=30, layer_slab=(2, 3))
    c = grating_coupler_circular(layer_slab=(3, 0))
    print(len(c.name))
    print(c.ports)
    c.show()



import pytest
from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
import gdsfactory.simulation as sim
from gdsfactory import components
from gdsfactory.simulation.read import read_sparameters_lumerical

factory = {
    i: getattr(components, i)
    for i in dir(components)
    if not i.startswith("_") and callable(getattr(components, i))
}


component_types = [
    "straight",
    "bend_circular",
    "bend_euler",
    "coupler",
    "mmi1x2",
    "mmi2x2",
]
component_types = []


@pytest.mark.parametrize("component_type", component_types)
def test_read_sparameters(
    component_type: str, data_regression: DataRegressionFixture, check: bool = True
) -> None:
    c = factory[component_type]()
    sp = sim.read_sparameters_lumerical(component=c)

    port_names = sp[0]
    f = list(sp[1])
    s = sp[2]

    lenf = s.shape[0]
    rows = s.shape[1]
    cols = s.shape[2]

    assert rows == cols == len(c.ports)
    assert len(port_names) == len(c.ports)
    if check:
        data_regression.check(dict(port_names=port_names))
    assert lenf == len(f)


def test_read_sparameters_2port_bend():
    filepath = gf.CONFIG["sparameters"] / "bend_circular" / "bend_circular_S220.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=2)
    assert port_names == ("N0", "W0"), port_names


def test_read_sparameters_2port_straight():
    filepath = gf.CONFIG["sparameters"] / "straight" / "straight_S220.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=2)
    assert len(f) == 500
    assert port_names == ("E0", "W0"), port_names


def test_read_sparameters_3port_mmi1x2():
    filepath = gf.CONFIG["sparameters"] / "mmi1x2" / "mmi1x2_si220n.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=3)
    assert len(f) == 500
    assert port_names == ("E0", "E1", "W0"), port_names


def test_read_sparameters_4port_mmi2x2():
    filepath = gf.CONFIG["sparameters"] / "mmi2x2" / "mmi2x2_si220n.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=4)
    assert len(f) == 500
    assert port_names == ("E0", "E1", "W0", "W1"), port_names


if __name__ == "__main__":
    # c = gf.components.straight(layer=(2, 0))
    # print(c.get_sparameters_path())
    test_read_sparameters("straight", None, False)



"""Write Sparameters with Lumerical FDTD."""

import dataclasses
import time
from pathlib import Path
from typing import Optional

import numpy as np
import omegaconf
import pandas as pd

import gdsfactory as gf
from gdsfactory.config import __version__, logger
from gdsfactory.simulation.get_sparameters_path import get_sparameters_path
from gdsfactory.tech import (
    LAYER_STACK,
    SIMULATION_SETTINGS,
    LayerStack,
    SimulationSettings,
)
from gdsfactory.types import ComponentOrFactory

run_false_warning = """
You have passed run=False to debug the simulation

run=False returns the simulation session for you to debug and make sure it is correct

To compute the Sparameters you need to pass run=True
"""

MATERIAL_NAME_TO_LUMERICAL = {
    "si": "Si (Silicon) - Palik",
    "sio2": "SiO2 (Glass) - Palik",
    "sin": "Si3N4 (Silicon Nitride) - Phillip",
}


def write_sparameters_lumerical(
    component: ComponentOrFactory,
    session: Optional[object] = None,
    run: bool = True,
    overwrite: bool = False,
    dirpath: Path = gf.CONFIG["sparameters"],
    layer_stack: LayerStack = LAYER_STACK,
    simulation_settings: SimulationSettings = SIMULATION_SETTINGS,
    **settings,
) -> pd.DataFrame:
    r"""Returns and writes component Sparameters using Lumerical FDTD.

    If simulation exists it returns the Sparameters directly unless overwrite=True
    which forces a re-run of the simulation


    Writes Sparameters both in .CSV and .DAT (interconnect format) as well as
    simulation settings in .YAML

    In the CSV format you can see `S12m` where `m` stands for magnitude
    and `S12a` where `a` stands for angle in radians

    Your components need to have ports, that will extend over the PML.

    .. image:: https://i.imgur.com/dHAzZRw.png

    For your Fab technology you can overwrite

    - Simulation Settings
    - dirpath
    - layerStack

    converts gdsfactory units (um) to Lumerical units (m)

    Disclaimer: This function tries to extract Sparameters automatically
    is hard to make a function that will fit all your possible simulation settings.
    You can use this function as an inspiration to create your own.

    Args:
        component: Component to simulate
        session: you can pass a session=lumapi.FDTD() or it will create one
        run: True runs Lumerical, False only draws simulation
        overwrite: run even if simulation results already exists
        dirpath: where to store the Sparameters
        layer_stack: layer_stack
        simulation_settings: dataclass with all simulation_settings
        settings: overwrite any simulation settings
            background_material: for the background
            port_margin: on both sides of the port width (um)
            port_height: port height (um)
            port_extension: port extension (um)
            mesh_accuracy: 2 (1: coarse, 2: fine, 3: superfine)
            zmargin: for the FDTD region (um)
            ymargin: for the FDTD region (um)
            xmargin: for the FDTD region (um)
            wavelength_start: 1.2 (um)
            wavelength_stop: 1.6 (um)
            wavelength_points: 500
            simulation_time: (s) related to max path length 3e8/2.4*10e-12*1e6 = 1.25mm
            simulation_temperature: in kelvin (default = 300)
            frequency_dependendent_profile: computes mode profiles for different wavelengths
            field_profile_samples: number of wavelengths to compute field profile


    .. code::

         top view
              ________________________________
             |                               |
             | xmargin                       | port_extension
             |<------>          port_margin ||<-->
          ___|___________          _________||___
             |           \        /          |
             |            \      /           |
             |             ======            |
             |            /      \           |
          ___|___________/        \__________|___
             |   |                           |
             |   |ymargin                    |
             |   |                           |
             |___|___________________________|

        side view
              ________________________________
             |                               |
             |                               |
             |                               |
             |ymargin                        |
             |<---> _____         _____      |
             |     |     |       |     |     |
             |     |     |       |     |     |
             |     |_____|       |_____|     |
             |       |                       |
             |       |                       |
             |       |zmargin                |
             |       |                       |
             |_______|_______________________|



    Return:
        Sparameters pandas DataFrame (wavelength_nm, S11m, S11a, S12a ...)
        suffix `a` for angle in radians and `m` for module

    """
    component = component() if callable(component) else component
    sim_settings = dataclasses.asdict(simulation_settings)

    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    layer_to_material = layer_stack.get_layer_to_material()

    if hasattr(component.info, "simulation_settings"):
        sim_settings.update(component.info.simulation_settings)
        logger.info(
            "Updating {component.name} sim settings {component.simulation_settings}"
        )
    for setting in settings.keys():
        if setting not in sim_settings:
            raise ValueError(
                f"Invalid setting `{setting}` not in ({list(sim_settings.keys())})"
            )

    sim_settings.update(**settings)
    ss = SimulationSettings(**sim_settings)

    component_extended = gf.c.extend_ports(
        component, length=ss.distance_source_to_monitors
    )

    ports = component_extended.get_ports_list(port_type="optical")
    if not ports:
        raise ValueError(f"`{component.name}` does not have any optical ports")

    c = gf.components.extension.extend_ports(
        component=component, length=ss.port_extension
    )
    c.remove_layers(component.layers - set(layer_to_thickness.keys()))
    c._bb_valid = False
    c.flatten()
    c.name = "top"
    gdspath = c.write_gds()

    filepath = get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_to_material=layer_to_material,
        layer_to_thickness=layer_to_thickness,
        **settings,
    )
    filepath_csv = filepath.with_suffix(".csv")
    filepath_sim_settings = filepath.with_suffix(".yml")
    filepath_fsp = filepath.with_suffix(".fsp")

    if run and filepath_csv.exists() and not overwrite:
        logger.info(f"Reading Sparameters from {filepath_csv}")
        return pd.read_csv(filepath_csv)

    if not run and session is None:
        print(run_false_warning)

    logger.info(f"Writing Sparameters to {filepath_csv}")
    x_min = (component.xmin - ss.xmargin) * 1e-6
    x_max = (component.xmax + ss.xmargin) * 1e-6
    y_min = (component.ymin - ss.ymargin) * 1e-6
    y_max = (component.ymax + ss.ymargin) * 1e-6

    layers_thickness = [
        layer_to_thickness[layer]
        for layer in component.get_layers()
        if layer in layer_to_thickness
    ]
    if not layers_thickness:
        raise ValueError(
            f"no layers for component {component.get_layers()}"
            f"in layer stack {layers_thickness.keys()}"
        )
    layers_zmin = [
        layer_to_zmin[layer]
        for layer in component.get_layers()
        if layer in layer_to_zmin
    ]
    component_thickness = max(layers_thickness)
    component_zmin = min(layers_zmin)

    z = (component_zmin + component_thickness) / 2 * 1e-6
    z_span = (2 * ss.zmargin + component_thickness) * 1e-6

    x_span = x_max - x_min
    y_span = y_max - y_min

    layers = c.get_layers()
    sim_settings.update(dict(layer_stack=layer_stack.to_dict()))

    sim_settings = dict(
        simulation_settings=sim_settings,
        component=component.to_dict(),
        version=__version__,
    )

    logger.info(
        f"Simulation size = {x_span*1e6:.3f}, {y_span*1e6:.3f}, {z_span*1e6:.3f} um"
    )

    # from pprint import pprint
    # filepath_sim_settings.write_text(omegaconf.OmegaConf.to_yaml(sim_settings))
    # print(filepath_sim_settings)
    # pprint(sim_settings)
    # return

    try:
        import lumapi
    except ModuleNotFoundError as e:
        print(
            "Cannot import lumapi (Python Lumerical API). "
            "You can add set the PYTHONPATH variable or add it with `sys.path.append()`"
        )
        raise e
    except OSError as e:
        raise e

    start = time.time()
    s = session or lumapi.FDTD(hide=False)
    s.newproject()
    s.selectall()
    s.deleteall()
    s.addrect(
        x_min=x_min,
        x_max=x_max,
        y_min=y_min,
        y_max=y_max,
        z=z,
        z_span=z_span,
        index=1.5,
        name="clad",
    )

    material = ss.background_material
    if material not in MATERIAL_NAME_TO_LUMERICAL:
        raise ValueError(f"{material} not in {list(MATERIAL_NAME_TO_LUMERICAL.keys())}")
    material = MATERIAL_NAME_TO_LUMERICAL[material]
    s.setnamed("clad", "material", material)

    s.addfdtd(
        dimension="3D",
        x_min=x_min,
        x_max=x_max,
        y_min=y_min,
        y_max=y_max,
        z=z,
        z_span=z_span,
        mesh_accuracy=ss.mesh_accuracy,
        use_early_shutoff=True,
        simulation_time=ss.simulation_time,
        simulation_temperature=ss.simulation_temperature,
    )

    for layer, thickness in layer_to_thickness.items():
        if layer not in layers:
            continue

        if layer not in layer_to_material:
            raise ValueError(f"{layer} not in {layer_to_material.keys()}")

        material_name = layer_to_material[layer]
        if material_name not in MATERIAL_NAME_TO_LUMERICAL:
            raise ValueError(
                f"{material_name} not in {list(MATERIAL_NAME_TO_LUMERICAL.keys())}"
            )
        material_name_lumerical = MATERIAL_NAME_TO_LUMERICAL[material_name]

        if layer not in layer_to_zmin:
            raise ValueError(f"{layer} not in {list(layer_to_zmin.keys())}")

        zmin = layer_to_zmin[layer]
        zmax = zmin + thickness
        z = (zmax + zmin) / 2

        s.gdsimport(str(gdspath), "top", f"{layer[0]}:{layer[1]}")
        layername = f"GDS_LAYER_{layer[0]}:{layer[1]}"
        s.setnamed(layername, "z", z * 1e-6)
        s.setnamed(layername, "z span", thickness * 1e-6)
        s.setnamed(layername, "material", material_name_lumerical)
        logger.info(f"adding {layer}, thickness = {thickness} um, zmin = {zmin} um ")

    for i, port in enumerate(ports):
        zmin = layer_to_zmin[port.layer]
        thickness = layer_to_thickness[port.layer]
        z = (zmin + thickness) / 2
        zspan = 2 * ss.port_margin + thickness

        s.addport()
        p = f"FDTD::ports::port {i+1}"
        s.setnamed(p, "x", port.x * 1e-6)
        s.setnamed(p, "y", port.y * 1e-6)
        s.setnamed(p, "z", z * 1e-6)
        s.setnamed(p, "z span", zspan * 1e-6)
        s.setnamed(p, "frequency dependent profile", ss.frequency_dependendent_profile)
        s.setnamed(p, "number of field profile samples", ss.field_profile_samples)

        deg = int(port.orientation)
        # if port.orientation not in [0, 90, 180, 270]:
        #     raise ValueError(f"{port.orientation} needs to be [0, 90, 180, 270]")

        if -45 <= deg <= 45:
            direction = "Backward"
            injection_axis = "x-axis"
            dxp = 0
            dyp = 2 * ss.port_margin + port.width
        elif 45 < deg < 90 + 45:
            direction = "Backward"
            injection_axis = "y-axis"
            dxp = 2 * ss.port_margin + port.width
            dyp = 0
        elif 90 + 45 < deg < 180 + 45:
            direction = "Forward"
            injection_axis = "x-axis"
            dxp = 0
            dyp = 2 * ss.port_margin + port.width
        elif 180 + 45 < deg < 180 + 45 + 90:
            direction = "Forward"
            injection_axis = "y-axis"
            dxp = 2 * ss.port_margin + port.width
            dyp = 0

        else:
            raise ValueError(
                f"port {port.name} orientation {port.orientation} is not valid"
            )

        s.setnamed(p, "direction", direction)
        s.setnamed(p, "injection axis", injection_axis)
        s.setnamed(p, "y span", dyp * 1e-6)
        s.setnamed(p, "x span", dxp * 1e-6)
        # s.setnamed(p, "theta", deg)
        s.setnamed(p, "name", port.name)
        # s.setnamed(p, "name", f"o{i+1}")

        logger.info(
            f"port {p} {port.name}: at ({port.x}, {port.y}, 0)"
            f"size = ({dxp}, {dyp}, {zspan})"
        )

    s.setglobalsource("wavelength start", ss.wavelength_start * 1e-6)
    s.setglobalsource("wavelength stop", ss.wavelength_stop * 1e-6)
    s.setnamed("FDTD::ports", "monitor frequency points", ss.wavelength_points)

    if run:
        s.save(str(filepath_fsp))
        s.deletesweep("s-parameter sweep")

        s.addsweep(3)
        s.setsweep("s-parameter sweep", "Excite all ports", 0)
        s.setsweep("S sweep", "auto symmetry", True)
        s.runsweep("s-parameter sweep")
        sp = s.getsweepresult("s-parameter sweep", "S parameters")
        s.exportsweep("s-parameter sweep", str(filepath))
        logger.info(f"wrote sparameters to {filepath}")

        keys = [key for key in sp.keys() if key.startswith("S")]
        ra = {f"{key}a": list(np.unwrap(np.angle(sp[key].flatten()))) for key in keys}
        rm = {f"{key}m": list(np.abs(sp[key].flatten())) for key in keys}
        wavelength_nm = sp["lambda"].flatten() * 1e9

        results = {"wavelength_nm": wavelength_nm}
        results.update(ra)
        results.update(rm)
        df = pd.DataFrame(results, index=wavelength_nm)

        end = time.time()
        df.to_csv(filepath_csv, index=False)
        sim_settings.update(compute_time_seconds=end - start)
        filepath_sim_settings.write_text(omegaconf.OmegaConf.to_yaml(sim_settings))
        return df
    filepath_sim_settings.write_text(omegaconf.OmegaConf.to_yaml(sim_settings))
    return s


def _sample_write_coupler_ring():
    """Write Sparameters when changing a component setting."""
    return [
        write_sparameters_lumerical(
            gf.components.coupler_ring(
                width=width, length_x=length_x, radius=radius, gap=gap
            )
        )
        for width in [0.5]
        for length_x in [0.1, 1, 2, 3, 4]
        for gap in [0.15, 0.2]
        for radius in [5, 10]
    ]


def _sample_bend_circular():
    """Write Sparameters for a circular bend with different radius."""
    return [
        write_sparameters_lumerical(gf.components.bend_circular(radius=radius))
        for radius in [2, 5, 10]
    ]


def _sample_bend_euler():
    """Write Sparameters for a euler bend with different radius."""
    return [
        write_sparameters_lumerical(gf.components.bend_euler(radius=radius))
        for radius in [2, 5, 10]
    ]


def _sample_convergence_mesh():
    return [
        write_sparameters_lumerical(
            component=gf.components.straight(length=2),
            mesh_accuracy=mesh_accuracy,
        )
        for mesh_accuracy in [1, 2, 3]
    ]


def _sample_convergence_wavelength():
    return [
        write_sparameters_lumerical(
            component=gf.components.straight(length=2),
            wavelength_start=wavelength_start,
        )
        for wavelength_start in [1.2, 1.4]
    ]


if __name__ == "__main__":
    # component = gf.components.straight(length=2.5)
    component = gf.components.mmi1x2()
    r = write_sparameters_lumerical(
        component=component, mesh_accuracy=1, wavelength_points=200, run=False
    )
    # c = gf.components.coupler_ring(length_x=3)
    # c = gf.components.mmi1x2()
    # print(r)
    # print(r.keys())
    # print(component.ports.keys())



import pathlib
from typing import List, Optional, Tuple

import numpy as np
from pydantic import validate_arguments

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import Port
from gdsfactory.tech import LAYER


@validate_arguments
def _add_pin_square(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.PORT,
    port_margin: float = 0.0,
) -> None:
    """Add half out pin to a component.

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label
        port_margin: margin to port edge


    .. code::

           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2 + port_margin

    dbot = np.array([pin_length / 2, -d])
    dtop = np.array([pin_length / 2, d])
    dbotin = np.array([-pin_length / 2, -d])
    dtopin = np.array([-pin_length / 2, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    if label_layer:
        component.add_label(
            text=str(p.name),
            position=p.midpoint,
            layer=label_layer,
        )


def _rotate(v, m):
    return np.dot(m, v)


def add_monitors_and_extend_ports(
    component: Component,
    port_labels: Optional[List[Port]] = None,
    extension_length: float = 0.2,
    port_margin: float = 1.0,
    layer: Tuple[int, int] = (101, 0),
) -> Component:
    """Add pins for port_labels
    each port increments the port layer

    Args:
        component: Component
        port_labels: list of labels
        extension_length: length to extend each port
        port_margin: margin from edge of the waveguide to edge of the port
        layer: monitor layer

    Returns: component with extended ports and monitors

    """

    port_labels = port_labels or component.ports.keys()

    for i, port_label in enumerate(port_labels):
        _add_pin_square(
            component=component,
            port=component.ports[port_label],
            pin_length=0,
            port_margin=port_margin,
            layer=(layer[0] + i, layer[1]),
        )

    return gf.components.extension.extend_ports(
        component=component, length=extension_length
    )


@cell
def add_monitors(
    component: Component,
    source_port_name: str = "o1",
    extension_length: float = 1.0,
    source_distance_to_monitors: float = 0.2,
    port_margin: float = 1.0,
    layer_monitor: Tuple[int, int] = (101, 0),
    layer_source: Tuple[int, int] = (110, 0),
    layer_simulation_region: Tuple[int, int] = (2, 0),
    top: float = 0.0,
    bottom: float = 0.0,
    right: float = 0.0,
    left: float = 0.0,
) -> Component:
    """Add monitors in layer_monitor, incrementing the layer number for each port
    Then extends all ports by source_distance_to_monitors
    Then add source in source_port_name
    Finally extends all ports by extension_length
    returns device centered at x=0, y=0

    Args:
        component: to add monitors
        source_port_name: name of the port to add the mode source
        extension_length: extension lenght when extending the ports
        port_margin: from waveguide edge to port edge
        layer_monitor: layer for the fist monitor
        layer_source: layer for the source
        layer_simulation_region: layer for the simulation region
        top: simulation region top padding
        bottom: simulation region south padding
        right: simulation region east padding
        left: simulation region west padding

    Returns: component with extended ports and monitors

    """
    c = gf.Component(f"{component.name}_monitors")

    # add monitors
    component_with_monitors = add_monitors_and_extend_ports(
        component=component.copy(),
        extension_length=source_distance_to_monitors,
        port_margin=port_margin,
        layer=layer_monitor,
    )

    # add source
    component_with_source = add_monitors_and_extend_ports(
        component=component_with_monitors.copy(),
        extension_length=extension_length,
        port_labels=[source_port_name],
        layer=layer_source,
    )

    # add simulation region
    component_with_padding = gf.add_padding_container(
        component=component_with_source,
        default=0,
        layers=[layer_simulation_region],
        top=top,
        bottom=bottom,
        right=right,
        left=left,
    )
    c.add(component_with_padding)
    c.ports = component_with_padding.ports
    return c


if __name__ == "__main__":
    gdspath = pathlib.Path.cwd() / "waveguide.gds"
    c = gf.components.bend_circular(radius=5)
    # c = gf.components.waveguide(length=2)
    # cm = extend_ports(component=c)
    cm = add_monitors(component=c)
    gf.show(cm)



"""Dummy fill to keep density constant."""
from typing import Optional, Union

from phidl.geometry import fill_rectangle as _fill_rectangle

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.read import from_phidl
from gdsfactory.types import Float2, Floats, Layers


@cell
def fill_rectangle(
    component: Component,
    fill_layers: Layers,
    fill_size: Float2 = (5.0, 5.0),
    avoid_layers: Optional[Layers] = None,
    include_layers: Optional[Layers] = None,
    margin: float = 5.0,
    fill_densities: Union[float, Floats] = (0.5, 0.25, 0.7),
    fill_inverted: bool = False,
    bbox: Optional[Float2] = None,
) -> Component:
    """Creates a rectangular fill pattern and fills all empty areas
    in the input component and returns a component that contains just the fill
    Dummy fill keeps density constant during fabrication

    Args:
        component: Component to fill
        fill_size: Rectangular size of the fill element
        avoid_layers: Layers to be avoided (not filled) in D
        include_layers: Layers to be filled, supercedes avoid_layers
        margin :
            Margin spacing around avoided areas -- fill will not come within
            `margin` of the geometry in D
        fill_layers: list of layers. fill pattern layers
        fill_densities: float between 0 and 1
            Defines the fill pattern density (1.0 == fully filled)
        fill_inverted: Inverts the fill pattern
        bbox: array-like[2][2]
            Limit the fill pattern to the area defined by this bounding box

    """

    component_filled = _fill_rectangle(
        component,
        fill_size=fill_size,
        avoid_layers=avoid_layers or "all",
        include_layers=include_layers,
        margin=margin,
        fill_layers=fill_layers,
        fill_densities=fill_densities,
        fill_inverted=fill_inverted,
        bbox=bbox,
    )
    return from_phidl(component_filled)


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight()
    c = gf.add_padding(c)
    c << fill_rectangle(
        c,
        fill_layers=((2, 0),),
        # fill_densities=(1.0,),
        fill_densities=1.0,
        avoid_layers=((1, 0),),
        # bbox=(100.0, 100.0),
    )
    c.show()



import re
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.simulation.get_sparameters_path import get_sparameters_path
from gdsfactory.tech import LAYER_STACK, LayerStack


def get_ports(line: str) -> Tuple[str, str]:
    """Returns 2 port labels strings from interconnect file."""
    line = line.replace('"', "")
    line = line.replace("(", "")
    line_fields = line.split(",")
    port1 = line_fields[0]
    port2 = line_fields[3]
    return port1, port2


def _read_sparameters_file(
    filepath, numports: int
) -> Tuple[Tuple[str, ...], np.array, np.ndarray]:
    r"""Returns Sparameters from Lumerical interconnect export file.

    Args:
        filepath: Sparameters filepath (interconnect format)
        numports: number of ports

    Returns:
        port_names: list of port labels
        F: frequency 1d np.array
        S: Sparameters np.ndarray matrix

    """
    F = []
    S = []
    port_names = []

    with open(filepath, "r") as fid:
        for _i in range(numports):
            port_line = fid.readline()
            m = re.search(r'\[".*",', port_line)
            if m:
                port = m.group(0)
                port_names.append(port[2:-2])
        line = fid.readline()
        port1, port2 = get_ports(line)
        line = fid.readline()
        numrows = int(tuple(line[1:-2].split(","))[0])
        S = np.zeros((numrows, numports, numports), dtype="complex128")
        r = m = n = 0
        for line in fid:
            if line[0] == "(":
                if "transmission" in line:
                    port1, port2 = get_ports(line)
                continue
            data = line.split()
            data = list(map(float, data))
            if m == 0 and n == 0:
                F.append(data[0])

            i = port_names.index(port1)
            j = port_names.index(port2)
            S[r, i, j] = data[1] * np.exp(1j * data[2])
            r += 1
            if r == numrows:
                r = 0
                m += 1
                if m == numports:
                    m = 0
                    n += 1
                    if n == numports:
                        break

    # port_names.reverse()
    # print(len(F), S.shape, len(port_names))
    return tuple(port_names), np.array(F), S


def read_sparameters_lumerical(
    component: Optional[Component] = None,
    layer_stack: LayerStack = LAYER_STACK,
    filepath: Optional[str] = None,
    numports: Optional[int] = None,
    dirpath: Path = gf.CONFIG["sparameters"],
) -> Tuple[List[str], np.array, np.ndarray]:
    r"""Returns Sparameters from Lumerical interconnect .DAT file.

    Args:
        component: Component
        layer_stack:
        filepath:
        numports: number of ports
        dirpath: path where to look for the Sparameters

    Returns:
        port_names: list of port labels
        F: frequency 1d np.array
        S: Sparameters np.ndarray matrix


    the Sparameters file have Lumerical format
    https://support.lumerical.com/hc/en-us/articles/360036107914-Optical-N-Port-S-Parameter-SPAR-INTERCONNECT-Element#toc_5
    """

    if component is None and filepath is None:
        raise ValueError("You need to define the filepath or the component")

    if filepath and numports is None:
        raise ValueError("You need to define numports")

    filepath = filepath or get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_to_material=layer_stack.get_layer_to_material(),
        layer_to_thickness=layer_stack.get_layer_to_thickness(),
    )
    numports = numports or len(component.ports)
    assert (
        filepath.exists()
    ), f"Sparameters for {component.name} not found in {filepath}"
    assert numports > 1, f"number of ports = {numports} and needs to be > 1"
    return _read_sparameters_file(filepath=filepath, numports=numports)


def read_sparameters_pandas(
    component: Component,
    layer_stack: LayerStack = LAYER_STACK,
    dirpath: Path = gf.CONFIG["sparameters"],
    **kwargs,
) -> pd.DataFrame:
    """Returns Sparameters in a pandas DataFrame.

    `S11m` `m` stands for module `S11a` `a` stands for angle (in radians)

    Args:
        component: Component
        layer_stack:
        dirpath: path where to look for the Sparameters
        kwargs: simulation settings

    """
    filepath = get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_to_material=layer_stack.get_layer_to_material(),
        layer_to_thickness=layer_stack.get_layer_to_thickness(),
        **kwargs,
    )
    df = pd.read_csv(filepath.with_suffix(".csv"))
    df.index = df["wavelength_nm"]
    return df


if __name__ == "__main__":
    s = read_sparameters_pandas(gf.components.mmi2x2())



from pathlib import Path
from typing import Tuple

import gdsfactory as gf
from gdsfactory.mask.merge_json import merge_json
from gdsfactory.mask.merge_markdown import merge_markdown
from gdsfactory.mask.merge_test_metadata import merge_test_metadata
from gdsfactory.mask.write_labels import write_labels


def merge_metadata(
    gdspath: Path,
    labels_prefix: str = "opt",
    label_layer: Tuple[int, int] = gf.LAYER.TEXT,
    **kwargs
) -> None:
    """Merges all JSON metadata into a big JSON.

    Args:
        gdspath: GDSpath
        labels_prefix
        label_layer: layer for the labels
    """
    mdpath = gdspath.with_suffix(".md")
    jsonpath = gdspath.with_suffix(".json")

    build_directory = gdspath.parent.parent
    doe_directory = build_directory / "sweep"

    write_labels(gdspath=gdspath, prefix=labels_prefix, label_layer=label_layer)

    merge_json(doe_directory=doe_directory, jsonpath=jsonpath, **kwargs)
    merge_markdown(reports_directory=doe_directory, mdpath=mdpath)
    merge_test_metadata(gdspath, labels_prefix=labels_prefix)


if __name__ == "__main__":

    gdspath = gf.CONFIG["samples_path"] / "mask" / "build" / "mask" / "mask.gds"
    print(gdspath)
    merge_metadata(gdspath)



import itertools
import multiprocessing
import os
import re
import shutil
import sys
import time
from glob import glob
from multiprocessing import Pool
from subprocess import PIPE, Popen, check_call

from gdsfactory.config import CONFIG, logger
from gdsfactory.read.from_yaml import factory as component_factory
from gdsfactory.sweep.read_sweep import read_sweep


def run_python(filename):
    """Run a python script and keep track of some context"""
    logger.debug("Running `{}`.".format(filename))
    command = ["python", filename]

    # Run the process
    t = time.time()
    process = Popen(command, stdout=PIPE, stderr=PIPE)
    stdout, _ = process.communicate()
    total_time = time.time() - t
    if process.returncode == 0:
        logger.info("v {} ({:.1f}s)".format(os.path.relpath(filename), total_time))
    else:
        logger.info(
            "! Error in {} {:.1f}s)".format(os.path.relpath(filename), total_time)
        )
        # message = "! Error in `{}`".format(basename(filename))
        # logger.error(message, exc_info=(Exception, stderr.strip(), None))
    if len(stdout.decode().strip()) > 0:
        logger.debug("Output of python {}:\n{}".format(filename, stdout.strip()))
    return filename, process.returncode


def build_devices(regex=".*", overwrite=True):
    """Builds all the python files in devices/"""
    # Avoid accidentally rebuilding devices
    if (
        os.path.isdir(CONFIG["gds_directory"])
        and os.listdir(CONFIG["gds_directory"])
        and not overwrite
    ):
        print("Run `make clean` to remove already built devices.")
        sys.exit(0)

    # Collect all the files to run.
    all_files = [
        os.path.join(dp, f)
        for dp, dn, filenames in os.walk(CONFIG["devices_directory"])
        for f in filenames
        if os.path.splitext(f)[1] == ".py"
    ]
    all_files = sorted(all_files)
    all_files = [f for f in all_files if re.search(regex, f)]

    # Notify user
    logger.info(
        "Building splits on {} threads. {} files to run.".format(
            multiprocessing.cpu_count(), len(all_files)
        )
    )
    logger.info(
        "Debug information at {}".format(
            os.path.relpath(os.path.join(CONFIG["log_directory"], "debug.log"))
        )
    )

    # Now run all the files in batches of $CPU_SIZE.
    with Pool(processes=multiprocessing.cpu_count()) as pool:
        for filename, rc in pool.imap_unordered(run_python, all_files):
            logger.debug("Finished {} {}".format(filename, rc))

    # Report on what we did.
    devices = glob(os.path.join(CONFIG["gds_directory"], "*.gds"))
    countmsg = "There are now {} GDS files in {}.".format(
        len(devices), os.path.relpath(CONFIG["gds_directory"])
    )
    logger.info(f"Finished building devices. {countmsg}")


def build_clean():
    """Cleans generated files such as build/."""
    target = CONFIG["build_directory"]
    if os.path.exists(target):
        shutil.rmtree(target)
        logger.info(f"Deleted {target}")


def build_cache_pull():
    """Pull devices from the cache"""
    if CONFIG.get("cache_url"):
        logger.info("Loading devices from cache...")
        check_call(
            [
                "rsync",
                "-rv",
                "--delete",
                CONFIG["cache_url"],
                CONFIG["build_directory"] + "/",
            ]
        )


def build_cache_push():
    """Push devices to the cache"""
    if not os.listdir(CONFIG["build_directory"]):
        logger.info("Nothing to push")
        return

    if CONFIG.get("cache_url"):
        logger.info("Uploading devices to cache...")
        check_call(
            [
                "rsync",
                "-rv",
                CONFIG["build_directory"] + "/",
                CONFIG["cache_url"],
                "--delete",
            ]
        )


def _build_doe(doe_name, config, component_factory=component_factory):
    from gdsfactory.sweep.write_sweep import write_sweep

    doe = config["does"][doe_name]
    component_type = doe.get("component")
    component_function = component_factory[component_type]
    write_sweep(
        component_type=component_function,
        doe_name=doe_name,
        do_permutations=doe.get("do_permutations", True),
        list_settings=doe.get("settings"),
        description=doe.get("description"),
        analysis=doe.get("analysis"),
        test=doe.get("test"),
        functions=doe.get("functions"),
    )


def build_does(filepath, component_factory=component_factory):
    """this function is depreacted

    Writes DOE settings from config.yml file and writes GDS into build_directory

    If you want to use cache use gf.generate_does instead

    Write For each DOE:

    - GDS
    - json metadata
    - ports CSV
    - markdown report, with DOE settings
    """

    does = read_sweep(filepath)
    doe_names = does.keys()

    doe_params = zip(
        doe_names, itertools.repeat(filepath), itertools.repeat(component_factory)
    )
    p = multiprocessing.Pool(multiprocessing.cpu_count())
    p.starmap(_build_doe, doe_params)

    # for doe_name in doe_names:
    #     p = multiprocessing.Process(target=_build_doe, args=(doe_name, config, component_factory=component_factory))
    #     p.start()


if __name__ == "__main__":
    does_path = CONFIG["samples_path"] / "mask" / "does.yml"
    build_does(does_path)

    # run_python("name.py")



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.component import Component


def test_get_bundle_udirect(
    data_regression: DataRegressionFixture, check: bool = True
) -> Component:

    c = gf.Component("test_get_bundle_udirect")

    pad = gf.partial(gf.c.pad, size=(10, 10))
    pad_south = gf.components.pad_array(orientation=270, spacing=(15.0, 0), pad=pad)
    pt = c << pad_south
    pb = c << pad_south
    pb.rotate(90)
    pt.rotate(90)
    pb.move((0, -100))

    pbports = pb.get_ports_list()
    ptports = pt.get_ports_list()

    pbports.reverse()

    routes = gf.routing.get_bundle(pbports, ptports, radius=5)

    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
    return c


if __name__ == "__main__":
    c = test_get_bundle_udirect(None, check=False)
    c.show()



"""Combine multiple JSONs into one."""

import dataclasses
import json
from typing import Any, Dict, Optional

from omegaconf.dictconfig import DictConfig

from gdsfactory.config import CONFIG, TECH, logger, write_config
from gdsfactory.types import PathType, PathTypes


def merge_json(
    doe_directory: PathType = CONFIG["doe_directory"],
    gds_directory: PathType = CONFIG["gds_directory"],
    extra_directories: Optional[PathTypes] = None,
    jsonpath: PathType = CONFIG["mask_directory"] / "metadata.json",
    json_version: int = 6,
    config: DictConfig = TECH,
) -> Dict[str, Any]:
    """Combine several JSON files from config.yml
    in the root of the mask directory, gets mask_name from there

    Args:
        doe_directory: defaults to current working directory
        extra_directories: list of extra_directories
        jsonpath
        json_version:
        config

    """
    logger.debug("Merging JSON files:")
    cells = {}
    extra_directories = extra_directories or []
    config = dataclasses.asdict(config)
    config.pop("library", "")

    for directory in extra_directories + [doe_directory]:
        for filename in directory.glob("*/*.json"):
            logger.debug(filename)
            with open(filename, "r") as f:
                data = json.load(f)
                cells.update(data.get("cells"))

    does = {d.stem: json.loads(open(d).read()) for d in doe_directory.glob("*.json")}
    metadata = dict(
        json_version=json_version,
        cells=cells,
        does=does,
        config=config,
    )

    write_config(metadata, jsonpath)
    logger.info(f"Wrote  metadata in {jsonpath}")
    return metadata


if __name__ == "__main__":
    from pprint import pprint

    d = merge_json()
    pprint(d)

    # print(config["module_versions"])
    # pprint(d['does'])

    # with open(jsonpath, "w") as f:
    #     f.write(json.dumps(d, indent=2))



import phidl.geometry as pg

import gdsfactory as gf


def component_phidl(function_name: str, **kwargs) -> gf.Component:
    if not hasattr(pg, function_name):
        raise ValueError(f"{function_name} not in {dir(pg)}")
    component_function = getattr(pg, function_name)
    device = component_function(**kwargs)
    return gf.read.from_phidl(device)


def test_import_component_phidl():
    component_phidl(function_name="snspd")


if __name__ == "__main__":
    test_import_component_phidl()



import json
import pathlib
import warnings
from functools import partial
from pathlib import Path
from typing import Callable, Optional, Tuple, Union, cast

import gdspy
import numpy as np
from phidl.device_layout import CellArray, DeviceReference

from gdsfactory.cell import cell_without_validator
from gdsfactory.component import Component
from gdsfactory.config import CONFIG
from gdsfactory.port import (
    Port,
    auto_rename_ports,
    read_port_markers,
    sort_ports_clockwise,
)
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import Layer, PathType


def add_ports_from_markers_square(
    component: Component,
    pin_layer: Layer = (69, 0),
    port_layer: Optional[Layer] = None,
    orientation: Optional[int] = 90,
    min_pin_area_um2: float = 0,
    max_pin_area_um2: float = 150 * 150,
    pin_extra_width: float = 0.0,
    port_names: Optional[Tuple[str, ...]] = None,
    port_name_prefix: str = "o",
) -> Component:
    """add ports from square markers at the port center in port_layer

    Args:
        component: to read polygons from and to write ports to
        pin_layer: for port markers
        port_layer: for the new created port
        orientation: in degrees 90: north, 0: east, 180: west, 270: south
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2
        max_pin_area_um2: ignore pins for area above certain size
        pin_extra_width: 2*offset from pin to straight
        port_names: names of the ports (defaults to {i})

    """
    port_markers = read_port_markers(component, [pin_layer])
    port_names = port_names or [
        f"{port_name_prefix}{i+1}" for i in range(len(port_markers.polygons))
    ]
    layer = port_layer or pin_layer

    for port_name, p in zip(port_names, port_markers.polygons):
        dy = snap_to_grid(p.ymax - p.ymin)
        dx = snap_to_grid(p.xmax - p.xmin)
        x = p.x
        y = p.y
        if dx == dy and max_pin_area_um2 > dx * dy > min_pin_area_um2:
            component.add_port(
                port_name,
                midpoint=(x, y),
                width=dx - pin_extra_width,
                orientation=orientation,
                layer=layer,
            )
    return component


def add_ports_from_markers_center(
    component: Component,
    pin_layer: Layer = (1, 10),
    port_layer: Optional[Layer] = None,
    inside: bool = False,
    tol: float = 0.1,
    pin_extra_width: float = 0.0,
    min_pin_area_um2: Optional[float] = None,
    max_pin_area_um2: float = 150.0 * 150.0,
    skip_square_ports: bool = True,
    xcenter: Optional[float] = None,
    ycenter: Optional[float] = None,
    port_name_prefix: str = "o",
    port_type: str = "optical",
    auto_rename_ports: bool = True,
) -> Component:
    """Add ports from rectangular pin markers.

    markers at port center, so half of the marker goes inside and half ouside the port.

    guess port orientation from the component center (xcenter)

    Args:
        component: to read polygons from and to write ports to
        pin_layer: GDS layer for maker [int, int]
        port_layer: for the new created port
        inside: True-> markers  inside. False-> markers at center
        tol: tolerance for comparing how rectangular is the pin
        pin_extra_width: 2*offset from pin to straight
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2
        max_pin_area_um2: ignore pins for area above certain size
        skip_square_ports: skips square ports (hard to guess orientation)
        xcenter: for guessing orientation of rectangular ports
        ycenter: for guessing orientation of rectangular ports
        port_name_prefix: o for optical ports (o1, o2, o3)
        port_type: type of port (optical, electrical ...)
        auto_rename_ports:

    For the default center case (inside=False)

    .. code::
           _______________
          |               |
          |               |
         |||             |||____  | pin_extra_width/2 > 0
         |||             |||
         |||             |||____
         |||             |||
          |      __       |
          |_______________|
                 __


    For the inside case (inside=True)

    .. code::
           _______________
          |               |
          |               |
          |               |
          | |             |
          | |             |
          |      __       |
          |               |
          |_______________|



    dx < dy: port is east or west
        x > xc: east
        x < xc: west

    dx > dy: port is north or south
        y > yc: north
        y < yc: south

    dx = dy
        x > xc: east
        x < xc: west

    """
    xc = xcenter or component.x
    yc = ycenter or component.y
    xmax = component.xmax
    xmin = component.xmin
    ymax = component.ymax
    ymin = component.ymin

    port_markers = read_port_markers(component, layers=(pin_layer,))
    layer = port_layer or pin_layer
    port_locations = []

    ports = {}

    for i, p in enumerate(port_markers.polygons):
        port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else i
        dy = p.ymax - p.ymin
        dx = p.xmax - p.xmin
        x = p.x
        y = p.y
        if min_pin_area_um2 and dx * dy < min_pin_area_um2:
            warnings.warn(f"skipping port with min_pin_area_um2 {dx * dy}")
            continue

        if max_pin_area_um2 and dx * dy > max_pin_area_um2:
            continue

        if skip_square_ports and snap_to_grid(dx) == snap_to_grid(dy):
            warnings.warn("skipping square port with no clear orientation")
            continue

        pxmax = p.xmax
        pxmin = p.xmin
        pymax = p.ymax
        pymin = p.ymin

        if dx < dy and x > xc:  # east
            orientation = 0
            width = dy
            if inside:
                x = p.xmax
        elif dx < dy and x < xc:  # west
            orientation = 180
            width = dy
            if inside:
                x = p.xmin
        elif dx > dy and y > yc:  # north
            orientation = 90
            width = dx
            if inside:
                y = p.ymax
        elif dx > dy and y < yc:  # south
            orientation = 270
            width = dx
            if inside:
                y = p.ymin

        # port markers have same width and height
        # check which edge (E, W, N, S) they are closer to
        elif pxmax > xmax - tol:  # east
            orientation = 0
            width = dy
            x = p.xmax
        elif pxmin < xmin + tol:  # west
            orientation = 180
            width = dy
            x = p.xmin
        elif pymax > ymax - tol:  # north
            orientation = 90
            width = dx
            y = p.ymax
        elif pymin < ymin + tol:  # south
            orientation = 270
            width = dx
            y = p.ymin

        x = snap_to_grid(x)
        y = snap_to_grid(y)
        width = np.round(width - pin_extra_width, 3)

        if (x, y) not in port_locations:
            port_locations.append((x, y))
            ports[port_name] = Port(
                name=port_name,
                midpoint=(x, y),
                width=width,
                orientation=orientation,
                layer=layer,
                port_type=port_type,
            )

    ports = sort_ports_clockwise(ports)

    for port_name, port in ports.items():
        if port_name not in component.ports:
            component.add_port(name=port_name, port=port)
    if auto_rename_ports:
        component.auto_rename_ports()
    return component


add_ports_from_markers_inside = partial(add_ports_from_markers_center, inside=True)


def add_ports_from_labels(
    component: Component,
    port_width: float,
    port_layer: Layer,
    xcenter: Optional[float] = None,
    port_name_prefix: str = "o",
    port_type: str = "optical",
) -> Component:
    """Add ports from labels.
    Assumes that all ports have a label at the port center.
    """
    xc = xcenter or component.x
    yc = component.y
    for i, label in enumerate(component.labels):
        x, y = label.position
        port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else i
        if x > xc:  # east
            orientation = 0
        elif x < xc:  # west
            orientation = 180
        elif y > yc:  # north
            orientation = 90
        elif y < yc:  # south
            orientation = 270

        component.add_port(
            name=port_name,
            midpoint=(x, y),
            width=port_width,
            orientation=orientation,
            port_type=port_type,
            layer=port_layer,
        )
    return component


def import_gds(
    gdspath: Union[str, Path],
    cellname: Optional[str] = None,
    flatten: bool = False,
    snap_to_grid_nm: Optional[int] = None,
    name: Optional[str] = None,
    decorator: Optional[Callable] = None,
    max_name_length: int = 32,
    **kwargs,
) -> Component:
    """Returns a Componenent from a GDS file.

    Adapted from phidl/geometry.py

    Args:
        gdspath: path of GDS file
        cellname: cell of the name to import (None) imports top cell
        flatten: if True returns flattened (no hierarchy)
        snap_to_grid_nm: snap to different nm grid (does not snap if False)
        name: Optional name
        decorator: function to apply over the imported gds
        max_name_length: can truncate the name of the cell before importing it
        **kwargs: component.info
    """
    gdspath = Path(gdspath)
    if not gdspath.exists():
        raise FileNotFoundError(f"No file {gdspath} found")
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(str(gdspath))
    top_level_cells = gdsii_lib.top_level()
    cellnames = [c.name for c in top_level_cells]

    if cellname is not None:
        if cellname not in gdsii_lib.cells:
            raise ValueError(
                f"cell {cellname} is not in file {gdspath} with cells {cellnames}"
            )
        topcell = gdsii_lib.cells[cellname]
    elif cellname is None and len(top_level_cells) == 1:
        topcell = top_level_cells[0]
    elif cellname is None and len(top_level_cells) > 1:
        raise ValueError(
            f"import_gds() There are multiple top-level cells in {gdspath}, "
            f"you must specify `cellname` to select of one of them among {cellnames}"
        )
    if flatten:
        component = Component(name=name or cellname or cellnames[0])
        polygons = topcell.get_polygons(by_spec=True)

        for layer_in_gds, polys in polygons.items():
            component.add_polygon(polys, layer=layer_in_gds)

    else:
        D_list = []
        c2dmap = {}
        for c in gdsii_lib.cells.values():
            D = Component(name=c.name)
            D.polygons = c.polygons
            D.references = c.references
            D.name = c.name
            for label in c.labels:
                rotation = label.rotation
                if rotation is None:
                    rotation = 0
                label_ref = D.add_label(
                    text=label.text,
                    position=np.asfarray(label.position),
                    magnification=label.magnification,
                    rotation=rotation * 180 / np.pi,
                    layer=(label.layer, label.texttype),
                )
                label_ref.anchor = label.anchor
            c2dmap.update({c: D})
            D_list += [D]

        for D in D_list:
            # First convert each reference so it points to the right Device
            converted_references = []
            for e in D.references:
                ref_device = c2dmap[e.ref_cell]
                if isinstance(e, gdspy.CellReference):
                    dr = DeviceReference(
                        device=ref_device,
                        origin=e.origin,
                        rotation=e.rotation,
                        magnification=e.magnification,
                        x_reflection=e.x_reflection,
                    )
                    dr.owner = D
                    converted_references.append(dr)
                elif isinstance(e, gdspy.CellArray):
                    dr = CellArray(
                        device=ref_device,
                        columns=e.columns,
                        rows=e.rows,
                        spacing=e.spacing,
                        origin=e.origin,
                        rotation=e.rotation,
                        magnification=e.magnification,
                        x_reflection=e.x_reflection,
                    )
                    dr.owner = D
                    converted_references.append(dr)
            D.references = converted_references

            # Next convert each Polygon
            # temp_polygons = list(D.polygons)
            # D.polygons = []
            # for p in temp_polygons:
            #     D.add_polygon(p)

            # Next convert each Polygon
            temp_polygons = list(D.polygons)
            D.polygons = []
            for p in temp_polygons:
                if snap_to_grid_nm:
                    points_on_grid = snap_to_grid(p.polygons[0], nm=snap_to_grid_nm)
                    p = gdspy.Polygon(
                        points_on_grid, layer=p.layers[0], datatype=p.datatypes[0]
                    )
                D.add_polygon(p)
        component = c2dmap[topcell]
        cast(Component, component)

    if decorator:
        component_new = decorator(component)
        component = component_new or component
    if flatten:
        component.flatten()

    component.info.update(**kwargs)

    component.name = name or component.name
    return cell_without_validator(lambda: component)(
        name=component.name, max_name_length=max_name_length, autoname=False
    )


def write_top_cells(gdspath: Union[str, Path], **kwargs) -> None:
    """Writes each top level cells into separate GDS file."""
    gdspath = pathlib.Path(gdspath)
    dirpath = gdspath.parent
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(gdspath)
    top_level_cells = gdsii_lib.top_level()
    cellnames = [c.name for c in top_level_cells]

    for cellname in cellnames:
        component = import_gds(gdspath, cellname=cellname, **kwargs)
        component.write_gds(f"{dirpath/cellname}.gds")


def write_cells(
    gdspath: Union[str, Path],
    dirpath: Optional[Union[str, Path]] = None,
) -> None:
    """Writes each top level cells into separate GDS file.

    Args:
        gdspath: to read cells from
        dirpath: directory path to store gds cells
    """
    gdspath = dirpath or pathlib.Path(gdspath)
    dirpath = dirpath or gdspath.parent / "gds"
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(gdspath)
    top_level_cells = gdsii_lib.top_level()
    cellnames = [c.name for c in top_level_cells]

    for cellname in cellnames:
        component = import_gds(gdspath, cellname=cellname)
        component.write_gds(f"{dirpath/cellname}.gds")
        write_cells_from_component(component=component, dirpath=dirpath)


def write_cells_from_component(
    component: Component, dirpath: Optional[PathType] = None
) -> None:
    """Writes all Component cells recursively.

    Args:
        component:
        dirpath: directory path to write GDS (defaults to CWD)
    """
    dirpath = dirpath or pathlib.Path(__file__).parent.absolute()
    if component.references:
        for ref in component.references:
            component = ref.parent
            component.write_gds(f"{dirpath/component.name}.gds")
            write_cells_from_component(component=component, dirpath=dirpath)
    else:
        component.write_gds(f"{dirpath/component.name}.gds")


def add_settings_from_label(component: Component) -> None:
    """Adds settings from label in JSON format in the GDS."""
    for label in component.labels:
        if label.text.startswith("settings="):
            d = json.loads(label.text[9:])
            component._settings_full = d.pop("settings", {})
            for k, v in d.items():
                setattr(component, k, v)


def _demo_optical() -> None:
    """Demo. See equivalent test in tests/import_gds_markers.py"""
    # c  =  gf.components.mmi1x2()
    # for p in c.ports.values():
    #     print(p)
    # c.show()

    gdspath = CONFIG["gdsdir"] / "mmi1x2.gds"
    c = import_gds(gdspath)
    add_ports_from_markers_center(c)
    auto_rename_ports(c)

    # for p in c.ports.values():
    #     print(p)


def _demo_electrical() -> None:
    """Demo. See equivalent test in tests/import_gds_markers.py"""

    gdspath = CONFIG["gdsdir"] / "mzi2x2.gds"
    c = import_gds(gdspath)
    add_ports_from_markers_center(c)
    auto_rename_ports(c)
    c.show()

    for p in c.ports.values():
        print(p)


def _demo_import_gds_markers() -> None:

    name = "mmi1x2"
    # name = "mzi2x2"
    gdspath = CONFIG["gdsdir"] / f"{name}.gds"
    c = import_gds(gdspath)
    add_ports_from_markers_center(c)
    assert len(c.ports) == 3
    return c


def test_import_gds_flat():
    gdspath = CONFIG["gdsdir"] / "mzi2x2.gds"
    c = import_gds(gdspath, snap_to_grid_nm=5, flatten=True)
    c.write_gds()


def test_import_gds():
    gdspath = CONFIG["gdsdir"] / "mzi2x2.gds"
    c = import_gds(gdspath, snap_to_grid_nm=5, flatten=False)
    c.write_gds()


if __name__ == "__main__":
    c = _demo_import_gds_markers()

    gdspath = CONFIG["gdsdir"] / "mzi2x2.gds"
    # c = import_gds(gdspath, snap_to_grid_nm=5, flatten=True, name="TOP")
    c = import_gds(gdspath, snap_to_grid_nm=5, flatten=True)
    print(c)
    c.show()



"""`get_route` returns a Manhattan route between two ports.

`get_route` only works for an individual routes. For routing groups of ports you need to use `get_bundle` instead

To make a route, you need to supply:

 - input port
 - output port
 - bend
 - straight
 - taper to taper to wider straights and reduce straight loss (Optional)


To generate a straight route:

 1. Generate the backbone of the route.
 This is a list of manhattan coordinates that the route would pass through
 if it used only sharp bends (right angles)

 2. Replace the corners by bend references
 (with rotation and position computed from the manhattan backbone)

 3. Add tapers if needed and if space permits

 4. generate straight portions in between tapers or bends


 A `Route` is a dataclass with:

- references: list of references for tapers, bends and straight waveguides
- ports: a dict of port name to Port, usually two ports "input" and "output"
- length: a float with the length of the route

"""
from functools import partial
from typing import Callable, Optional

import numpy as np

from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.components.wire import wire_corner
from gdsfactory.cross_section import metal3, strip
from gdsfactory.port import Port
from gdsfactory.routing.manhattan import round_corners, route_manhattan
from gdsfactory.types import (
    ComponentFactory,
    ComponentOrFactory,
    Coordinates,
    CrossSectionFactory,
    Route,
)


def get_route(
    input_port: Port,
    output_port: Port,
    bend: ComponentOrFactory = bend_euler,
    straight: ComponentOrFactory = straight_function,
    taper: Optional[ComponentFactory] = None,
    start_straight_length: float = 0.01,
    end_straight_length: float = 0.01,
    min_straight_length: float = 0.01,
    cross_section: CrossSectionFactory = strip,
    **kwargs,
) -> Route:
    """Returns a Manhattan Route between 2 ports
    The references are straights, bends and tapers.
    `get_route` is an automatic version of `get_route_from_steps`

    Args:
        input_port: start port
        output_port: end port
        bend: function that return bends
        straight: function that returns straights
        taper:
        start_straight_length: length of starting straight
        end_straight_length: length of end straight
        min_straight_length: min length of straight for any intermediate segment
        cross_section:
        kwargs: cross_section settings


    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component('sample_connect')
        mmi1 = c << gf.components.mmi1x2()
        mmi2 = c << gf.components.mmi1x2()
        mmi2.move((40, 20))
        route = gf.routing.get_route(mmi1.ports["o2"], mmi2.ports["o1"], radius=5)
        c.add(route.references)
        c.plot()

    """
    x = cross_section(**kwargs)
    taper_length = x.info.get("taper_length")
    width1 = input_port.width
    auto_widen = x.info.get("auto_widen", False)
    width2 = x.info.get("width_wide") if auto_widen else width1

    bend90 = bend(cross_section=cross_section, **kwargs) if callable(bend) else bend

    if taper:
        taper = partial(
            taper,
            length=taper_length,
            width1=input_port.width,
            width2=width2,
            cross_section=cross_section,
            **kwargs,
        )

    return route_manhattan(
        input_port=input_port,
        output_port=output_port,
        straight=straight,
        taper=taper,
        start_straight_length=start_straight_length,
        end_straight_length=end_straight_length,
        min_straight_length=min_straight_length,
        bend=bend90,
        cross_section=cross_section,
        **kwargs,
    )


get_route_electrical = partial(
    get_route,
    bend=wire_corner,
    start_straight_length=10,
    end_straight_length=10,
    cross_section=metal3,
    taper=None,
)


def get_route_from_waypoints(
    waypoints: Coordinates,
    bend: Callable = bend_euler,
    straight: Callable = straight_function,
    taper: Optional[Callable] = taper_function,
    cross_section: CrossSectionFactory = strip,
    **kwargs,
) -> Route:
    """Returns a route formed by the given waypoints with
    bends instead of corners and optionally tapers in straight sections.
    Tapering to wider straights reduces the optical loss.
    `get_route_from_waypoints` is a manual version of `get_route`
    `get_route_from_steps` is a  more concise and convenient version of
    `get_route_from_waypoints` also available in gf.routing

    Args:
        waypoints: Coordinates that define the route
        bend: function that returns bends
        straight: function that returns straight waveguides
        taper: function that returns tapers
        cross_section:
        kwargs: cross_section settings

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component("waypoints_sample")

        w = gf.components.straight()
        left = c << w
        right = c << w
        right.move((100, 80))

        obstacle = gf.components.rectangle(size=(100, 10))
        obstacle1 = c << obstacle
        obstacle2 = c << obstacle
        obstacle1.ymin = 40
        obstacle2.xmin = 25


        p0x, p0y = left.ports["o2"].midpoint
        p1x, p1y = right.ports["o2"].midpoint
        o = 10  # vertical offset to overcome bottom obstacle
        ytop = 20


        routes = gf.routing.get_route_from_waypoints(
            [
                (p0x, p0y),
                (p0x + o, p0y),
                (p0x + o, ytop),
                (p1x + o, ytop),
                (p1x + o, p1y),
                (p1x, p1y),
            ],
        )
        c.add(routes.references)
        c.plot()

    """

    x = cross_section(**kwargs)
    auto_widen = x.info.get("auto_widen", False)
    width1 = x.info.get("width")
    width2 = x.info.get("width_wide") if auto_widen else width1
    taper_length = x.info.get("taper_length")
    waypoints = np.array(waypoints)
    kwargs.pop("route_filter", "")

    if auto_widen:
        taper = (
            taper(
                length=taper_length,
                width1=width1,
                width2=width2,
                cross_section=cross_section,
                **kwargs,
            )
            if callable(taper)
            else taper
        )
    else:
        taper = None

    return round_corners(
        points=waypoints,
        bend=bend,
        straight=straight,
        taper=taper,
        cross_section=cross_section,
        **kwargs,
    )


if __name__ == "__main__":
    import gdsfactory as gf

    # w = gf.components.mmi1x2()
    # c = gf.Component()
    # c << w
    # route = get_route(w.ports["o2"], w.ports["o1"], layer=(2, 0), width=2)
    # cc = c.add(route.references)
    # cc.show()

    c = gf.Component()
    p1 = c << gf.c.pad_array270()
    p2 = c << gf.c.pad_array90()

    p1.movex(300)
    p1.movey(300)
    route = get_route_electrical(
        p1.ports["e13"],
        p2.ports["e11"],
        cross_section=gf.cross_section.metal3,
        width=10.0,
    )
    c.add(route.references)

    c.show()



# import matplotlib.pyplot as plt
import meep as mp


def plot_xsection(sim, center=(0, 0, 0), size=(0, 2, 2)):
    """
    sim: simulation object
    """
    sim.plot2D(output_plane=mp.Volume(center=center, size=size))
    # plt.colorbar()



from typing import Optional, Tuple

from numpy import floor

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.via import via1, viac
from gdsfactory.tech import LAYER
from gdsfactory.types import ComponentOrFactory, Layer


@gf.cell
def contact_slot(
    size: Tuple[float, float] = (11.0, 11.0),
    layers: Tuple[Layer, ...] = (LAYER.M1, LAYER.M2),
    layer_offsets: Tuple[float, ...] = (0, 1.0),
    layer_port: Optional[Layer] = None,
    via: ComponentOrFactory = via1,
    enclosure: float = 1.0,
    ysize: float = 0.5,
    yspacing: float = 2.0,
) -> Component:
    """Rectangular contact with slotted via in X direction

    Args:
        size: of the layers
        layers: layers on which to draw rectangles
        layer_offsets: cladding_offset for each layer
        layer_port: if None asumes port is on the last layer
        via: via to use to fill the rectangles
        enclosure: of the via by rectangle
        ysize: via height in y
        yspacing: via spacing pitch in y

    .. code::

        enclosure
        _____________________________________
        |<--->                              |
        |      ______________________       |
        |     |                      |      |
        |     |                      | ysize|
        |     |______________________|      |
        |  |                                |
        |  | yspacing                       |
        |  |                                |  size[1]
        |  |   ______________________       |
        |  |  |                      |      |
        |  |  |                      | ysize|
        |  |  |______________________|      |
        |                                   |
        |___________________________________|
                        size[0]


    """
    if size[0] - 2 * enclosure < 0:
        raise ValueError(
            f"Contact length (size[0] = {size[0]}) < 2*enclosure ({2*enclosure}). "
        )
    if size[1] - 2 * enclosure < 0:
        raise ValueError(
            f"Contact width (size[1] = {size[1]}) < 2*enclosure ({2*enclosure}). "
        )

    layer_port = layer_port or layers[-1]

    c = Component()
    c.info.size = (float(size[0]), float(size[1]))

    for layer, offset in zip(layers, list(layer_offsets) + [0] * len(layers)):
        ref = c << compass(
            size=(size[0] + 2 * offset, size[1] + 2 * offset), layer=layer
        )

        if layer == layer_port:
            c.add_ports(ref.ports)

    via = via(size=(size[0] - 2 * enclosure, ysize)) if callable(via) else via

    nb_vias_y = (size[1] - 2 * enclosure) / yspacing
    nb_vias_y = int(floor(nb_vias_y)) or 1
    ref = c.add_array(via, columns=1, rows=nb_vias_y, spacing=(0, yspacing))
    dy = (size[1] - (nb_vias_y - 1) * yspacing - size[1]) / 2
    ref.move((0, dy))
    return c


contact_slot_m1_m2 = gf.partial(contact_slot, layers=(LAYER.M1, LAYER.M2), via=via1)

contact_slot_slab_m1 = gf.partial(contact_slot, layers=(LAYER.M1,), via=viac)


if __name__ == "__main__":
    c = contact_slot_m1_m2(layer_offsets=(0.5, 1), enclosure=2, size=(1, 1))
    c.show()



from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.manhattan_font import manhattan_text
from gdsfactory.types import Coordinate, Layer


@gf.cell
def text_rectangular(
    text: str = "abcd",
    size: float = 10.0,
    position: Coordinate = (0, 0),
    justify: str = "left",
    layers: Tuple[Layer, ...] = ((1, 0),),
) -> Component:
    """Returns Pixel based rectangular text.

    Args:
        text: text
        size: pixel size
        position: coordinate
        justify: left, right, center
        layers: layers
    """
    c = gf.Component()
    for layer in layers:
        c.add_ref(
            manhattan_text(
                text=text, size=size, position=position, justify=justify, layer=layer
            )
        )
    return c


if __name__ == "__main__":
    c = text_rectangular(
        text=".[,ABCDEFGHIKKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789:/",
        size=4.0,
        justify="right",
        position=(120.5, 3),
        layers=[(1, 0), (2, 0)],
    )
    c.show()



"""Returns tidy3d simulation from gdsfactory Component."""
import warnings
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pydantic
import tidy3d as td

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.extension import move_polar_rad_copy
from gdsfactory.routing.sort_ports import sort_ports_x, sort_ports_y
from gdsfactory.simulation.tidy3d.materials import get_material
from gdsfactory.tech import LAYER_STACK, LayerStack

MATERIAL_NAME_TO_TIDY3D = {
    "si": "cSi",
    "sio2": "SiO2",
    "sin": "Si3N4",
}


@pydantic.validate_arguments
def get_simulation(
    component: Component,
    mode_index: int = 0,
    n_modes: int = 2,
    port_extension: Optional[float] = 4.0,
    layer_stack: LayerStack = LAYER_STACK,
    zmargin: float = 1.0,
    thickness_pml: float = 1.0,
    clad_material: str = "SiO2",
    port_source_name: str = "o1",
    port_margin: float = 0.5,
    distance_source_to_monitors: float = 0.2,
    mesh_step: float = 40e-3,
    wavelength: float = 1.55,
) -> td.Simulation:
    """Returns Simulation object from gdsfactory.component

    based on GDS example
    https://simulation.cloud/docs/html/examples/ParameterScan.html

    Args:
        component: gf.Component
        mode_index: mode index
        n_modes: number of modes
        port_extension: extend ports beyond the PML
        layer_stack: contains layer numbers (int, int) to thickness, zmin
        zmargin: thickness for cladding above and below core
        thickness_pml: PML thickness (um)
        clad_material: material for cladding
        port_source_name: input port name
        port_margin: margin on each side of the port
        distance_source_to_monitors: in (um) source goes before monitors
        mesh_step: in all directions
        wavelength: in (um)

    You can visualize the simulation with gdsfactory


    .. code::

        import matplotlib.pyplot as plt
        import gdsfactory as gf
        import gdsfactory.simulation.tidy3d as gm

        c = gf.components.bend_circular()
        sim = gm.get_simulation(c)
        gm.plot_simulation(sim)

    """
    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_material = layer_stack.get_layer_to_material()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    # layer_to_sidewall_angle = layer_stack.get_layer_to_sidewall_angle()

    assert isinstance(
        component, Component
    ), f"component needs to be a gf.Component, got Type {type(component)}"
    if port_source_name not in component.ports:
        warnings.warn(
            f"port_source_name={port_source_name} not in {component.ports.keys()}"
        )
        port_source = component.get_ports_list()[0]
        port_source_name = port_source.name
        warnings.warn(f"Selecting port_source_name={port_source_name} instead.")

    component_extended = (
        gf.components.extension.extend_ports(
            component=component, length=port_extension, centered=True
        )
        if port_extension
        else component
    )

    gf.show(component_extended)
    component_extended.flatten()
    component_extended_ref = component_extended.ref()

    component_ref = component.ref()
    component_ref.x = 0
    component_ref.y = 0

    structures = [
        td.Box(
            material=get_material(name=clad_material),
            size=(td.inf, td.inf, td.inf),
            center=(0, 0, 0),
        )
    ]
    layers_thickness = [
        layer_to_thickness[layer]
        for layer in component.get_layers()
        if layer in layer_to_thickness
    ]

    t_core = max(layers_thickness)
    cell_thickness = thickness_pml + t_core + thickness_pml + 2 * zmargin
    sim_size = [
        component_ref.xsize + 2 * thickness_pml,
        component_ref.ysize + 2 * thickness_pml,
        cell_thickness,
    ]

    for layer in component.layers:
        if layer in layer_to_thickness and layer in layer_to_material:
            height = layer_to_thickness[layer]
            zmin = layer_to_zmin[layer]
            z_cent = zmin + height / 2
            material_name = MATERIAL_NAME_TO_TIDY3D[layer_to_material[layer]]
            material = get_material(name=material_name)

            geometry = td.GdsSlab(
                material=material,
                gds_cell=component_extended_ref,
                gds_layer=layer[0],
                gds_dtype=layer[1],
                z_cent=z_cent,
                z_size=height,
            )
            structures.append(geometry)

    # Add source
    port = component_ref.ports[port_source_name]
    angle = port.orientation
    width = port.width + 2 * port_margin
    size_x = width * abs(np.sin(angle * np.pi / 180))
    size_y = width * abs(np.cos(angle * np.pi / 180))
    size_x = 0 if size_x < 0.001 else size_x
    size_y = 0 if size_y < 0.001 else size_y
    size_z = cell_thickness - 2 * thickness_pml
    size = [size_x, size_y, size_z]
    center = port.center.tolist() + [0]  # (x, y, z=0)
    freq0 = td.constants.C_0 / wavelength
    fwidth = freq0 / 10

    msource = td.ModeSource(
        size=size,
        center=center,
        source_time=td.GaussianPulse(frequency=freq0, fwidth=fwidth),
        direction="forward",
    )

    # Add port monitors
    monitors = {}
    ports = sort_ports_x(sort_ports_y(component_ref.get_ports_list()))
    for port in ports:
        port_name = port.name
        angle = port.orientation
        width = port.width + 2 * port_margin
        size_x = width * abs(np.sin(angle * np.pi / 180))
        size_y = width * abs(np.cos(angle * np.pi / 180))
        size_x = 0 if size_x < 0.001 else size_x
        size_y = 0 if size_y < 0.001 else size_y
        size = (size_x, size_y, size_z)

        # if monitor has a source move monitor inwards
        length = -distance_source_to_monitors if port_name == port_source_name else 0
        xy_shifted = move_polar_rad_copy(
            np.array(port.center), angle=angle * np.pi / 180, length=length
        )
        center = xy_shifted.tolist() + [0]  # (x, y, z=0)

        monitors[port_name] = td.ModeMonitor(
            center=[port.x, port.y, t_core / 2],
            size=size,
            freqs=[freq0],
            Nmodes=1,
            name=port.name,
        )

    domain_monitor = td.FreqMonitor(
        center=[0, 0, z_cent], size=[sim_size[0], sim_size[1], 0], freqs=[freq0]
    )

    sim = td.Simulation(
        size=sim_size,
        mesh_step=mesh_step,
        structures=structures,
        sources=[msource],
        monitors=[domain_monitor] + list(monitors.values()),
        run_time=20 / fwidth,
        pml_layers=[12, 12, 12],
    )
    # set the modes
    sim.compute_modes(msource, Nmodes=n_modes)
    sim.set_mode(msource, mode_ind=mode_index)
    return sim


def plot_simulation(
    sim: td.Simulation,
    normal1: str = "z",
    normal2: str = "x",
    position1: float = 0.0,
    position2: float = 0.0,
):
    """Returns figure with two axis of the Simulation.

    Args:
        sim: simulation object
        normal1: {'x', 'y', 'z'} Axis normal to the cross-section plane.
        normal2: {'x', 'y', 'z'} Axis normal to the cross-section plane.
        position1: Position offset along the normal axis.
        position2: Position offset along the normal axis.

    """
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(14, 4))
    sim.viz_eps_2D(normal=normal1, position=position1, ax=ax1)
    sim.viz_eps_2D(normal=normal2, position=position2, ax=ax2, source_alpha=1)
    plt.show()
    return fig


def plot_materials(
    sim: td.Simulation,
    normal1: str = "z",
    normal2: str = "x",
    position1: float = 0.0,
    position2: float = 0.0,
):
    """Returns figure with two axis of the Simulation.

    Args:
        sim: simulation object
        normal1: {'x', 'y', 'z'} Axis normal to the cross-section plane.
        normal2: {'x', 'y', 'z'} Axis normal to the cross-section plane.
        position1: Position offset along the normal axis.
        position2: Position offset along the normal axis.

    """
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(14, 4))
    sim.viz_mat_2D(normal=normal1, position=position1, ax=ax1)
    sim.viz_mat_2D(
        normal=normal2, position=position2, ax=ax2, source_alpha=1, legend=True
    )
    plt.show()
    return fig


if __name__ == "__main__":
    c = gf.components.mmi1x2()
    c = gf.add_padding(c, default=0, bottom=2, top=2, layers=[(100, 0)])

    # c = gf.add_padding(c, default=0, bottom=2, right=2, layers=[(100, 0)])

    # c = gf.add_padding(c, default=0, bottom=2, top=2, layers=[(100, 0)])
    # c = gf.components.straight(length=2)

    c = gf.components.bend_circular(radius=2)

    sim = get_simulation(c)
    # fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(14, 4))
    # sim.viz_eps_2D(normal="z", position=0, ax=ax1)
    # sim.viz_eps_2D(normal="x", ax=ax2, source_alpha=1)
    # ax2.set_xlim([-3, 3])
    plot_simulation(sim)



import pathlib
from typing import Union

from gdsfactory import klive
from gdsfactory.cell import clear_cache as clear_cache_function
from gdsfactory.component import Component


def show(
    component: Union[Component, str, pathlib.Path], clear_cache: bool = True, **kwargs
) -> None:
    """Shows Component in klayout

    Args:
        component
        clear_cache: clear_cache
        kwargs: settings for write_gds
    """
    if isinstance(component, pathlib.Path):
        component = str(component)
        return klive.show(component)
    elif isinstance(component, str):
        return klive.show(component)
    elif component is None:
        raise ValueError(
            "Component is None, make sure that your function returns the component"
        )

    elif isinstance(component, Component):
        gdspath = component.write_gds(**kwargs)
        klive.show(gdspath)
    else:
        raise ValueError(
            f"Component is {type(component)}, make sure pass a Component or a path"
        )
    if clear_cache:
        clear_cache_function()



import inspect
import pathlib

from gdsfactory import components
from gdsfactory.component import Component


def write_factory(module=components, filepath="write_factory.pyc"):
    """Write a file with all the component factories into a dict."""
    factory = [
        i
        for i in dir(module)
        if not i.startswith("_")
        and callable(getattr(components, i))
        and type(inspect.signature(getattr(components, i)).return_annotation)
        == type(Component)
    ]

    lines = [f"{i}={i}" for i in factory]
    script = "factory = dict(" + ",".join(lines) + ")"
    filepath = pathlib.Path(filepath)
    print(script)
    filepath.write_text(script)


if __name__ == "__main__":
    write_factory()



import warnings
from typing import Optional, Tuple, Union

import numpy as np
from numpy import ndarray

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.cross_section import strip
from gdsfactory.port import Port
from gdsfactory.types import (
    ComponentOrFactory,
    Coordinate,
    CrossSectionOrFactory,
    Layer,
)

DEG2RAD = np.pi / 180


def line(
    p_start: Union[Port, Coordinate],
    p_end: Union[Port, Coordinate],
    width: Optional[float] = None,
) -> Tuple[float, float, float, float]:
    if isinstance(p_start, gf.Port):
        width = p_start.width
        p_start = p_start.midpoint

    if isinstance(p_end, gf.Port):
        p_end = p_end.midpoint

    w = width
    angle = np.arctan2(p_end[1] - p_start[1], p_end[0] - p_start[0])
    a = np.pi / 2
    p0 = move_polar_rad_copy(p_start, angle + a, w / 2)
    p1 = move_polar_rad_copy(p_start, angle - a, w / 2)
    p2 = move_polar_rad_copy(p_end, angle - a, w / 2)
    p3 = move_polar_rad_copy(p_end, angle + a, w / 2)
    return p0, p1, p2, p3


def move_polar_rad_copy(pos: Coordinate, angle: float, length: float) -> ndarray:
    """Returns the points of a position (pos) with angle, by shifted by certain length

    Args:
        pos: position
        angle: in radians
        length: extension length

    """
    c = np.cos(angle)
    s = np.sin(angle)
    return pos + length * np.array([c, s])


@cell
def extend_port(port: Port, length: float, layer: Optional[Layer] = None) -> Component:
    """Returns a straight extension component out of a port.

    Args:
        port: port to extend
        length: extension length
        layer: for the straight section
    """
    c = Component()
    layer = layer or port.layer

    # Generate a port extension
    p_start = port.midpoint
    angle = port.angle
    p_end = move_polar_rad_copy(p_start, angle * DEG2RAD, length)
    w = port.width

    _line = line(p_start, p_end, w)

    c.add_polygon(_line, layer=layer)
    c.add_port(name="original", port=port)

    port_settings = port.settings.copy()
    port_settings.update(midpoint=p_end)
    c.add_port(**port_settings)

    return c


@gf.cell
def extend_ports(
    component: ComponentOrFactory = mmi1x2,
    port_names: Optional[Tuple[str, ...]] = None,
    length: float = 5.0,
    extension_factory: Optional[ComponentOrFactory] = None,
    port1: Optional[str] = None,
    port2: Optional[str] = None,
    port_type: str = "optical",
    centered: bool = False,
    cross_section: CrossSectionOrFactory = strip,
    **kwargs,
) -> Component:
    """Returns a new component with some ports extended
    it can accept an extension_factory or it defaults to the port
    width and layer of each extended port

    Args:
        component: component to extend ports
        port_names: specify an list of ports names, if None it extends all ports
        length: extension length
        extension_factory: function to extend ports (defaults to a straight)
        port1: input port name
        port2: output port name
        port_type: type of the ports to extend
        centered: if True centers rectangle at (0, 0)
        kwargs: for selecting the ports
    """
    c = gf.Component()
    component = component() if callable(component) else component
    cref = c << component
    c.component = component

    if centered:
        cref.x = 0
        cref.y = 0

    ports_all = cref.get_ports_list()
    port_all_names = [p.name for p in ports_all]

    ports_to_extend = cref.get_ports_list(port_type=port_type, **kwargs)
    ports_to_extend_names = [p.name for p in ports_to_extend]
    port_names = port_names or ports_to_extend_names or port_all_names

    for port_name in port_names:
        if port_name not in port_all_names:
            warnings.warn(f"Port Name {port_name} not in {port_all_names}")

    for port in ports_all:
        port_name = port.name
        port = cref.ports[port_name]
        cross_section = port.cross_section or cross_section

        if port_name in port_names:
            if extension_factory:
                extension_component = extension_factory()
            else:
                extension_component = gf.partial(
                    gf.c.straight,
                    length=length,
                    width=port.width,
                    cross_section=cross_section,
                    layer=port.layer,
                )()
            port_labels = list(extension_component.ports.keys())
            port1 = port1 or port_labels[0]
            port2 = port2 or port_labels[-1]

            extension = c << extension_component
            extension.connect(port1, port)
            c.add_port(port_name, port=extension.ports[port2])
            c.absorb(extension)
        else:
            c.add_port(port_name, port=component.ports[port_name])

    c.copy_child_info(component)
    return c


def test_extend_ports() -> Component:
    import gdsfactory.components as pc

    c = pc.cross(width=2, port_type="optical")
    ce = extend_ports(component=c)
    assert len(c.ports) == len(ce.ports)
    p = len(ce.polygons)
    assert p == 4, p
    return ce


def test_extend_ports_selection() -> Component:
    import gdsfactory.components as pc

    c = pc.cross(width=2)
    ce = extend_ports(component=c, port_names=list(range(1, 4)))
    assert len(c.ports) == len(ce.ports)
    return ce


__all__ = ["extend_ports", "extend_port"]


if __name__ == "__main__":
    # c = extend_ports()
    # c = test_extend_ports_selection()
    # c = test_extend_ports()
    # c.show()

    # c = gf.c.bend_circular()
    # ce = extend_ports(component=c, port_names=list(c.ports.keys()) + ["hi"])
    # ce.show()

    c = gf.components.straight_pin(length=40)
    ce = extend_ports(c, port_names=("top_e1", "bot_e3"))
    # ce = extend_ports(c, port_names=("bad", "worse"))
    ce.show()
    # wg_pin.show()

    # c = pc.straight(layer=(3, 0))
    # print(ce)
    # print(len(ce.ports))
    # c = pc.straight()
    # ce = extend_ports(component=c)
    # ce.show()



"""Straight Doped PIN waveguide."""
from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.contact import contact_m1_m3
from gdsfactory.components.contact_slot import contact_slot_slab_m1
from gdsfactory.components.taper import taper_strip_to_ridge
from gdsfactory.cross_section import pin, pn
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def straight_pin_slot(
    length: float = 500.0,
    cross_section: CrossSectionFactory = pin,
    contact: ComponentFactory = contact_m1_m3,
    contact_width: float = 10.0,
    contact_slab: ComponentFactory = contact_slot_slab_m1,
    contact_slab_top: Optional[ComponentFactory] = None,
    contact_slab_bot: Optional[ComponentFactory] = None,
    contact_slab_width: Optional[float] = None,
    contact_spacing: float = 3.0,
    contact_slab_spacing: float = 2.0,
    taper: Optional[ComponentFactory] = taper_strip_to_ridge,
    **kwargs,
) -> Component:
    """Returns a PIN straight waveguide with slotted via

    https://doi.org/10.1364/OE.26.029983

    500um length for PI phase shift
    https://ieeexplore.ieee.org/document/8268112

    to go beyond 2PI, you will need at least 1mm
    https://ieeexplore.ieee.org/document/8853396/

    Args:
        length: of the waveguide
        cross_section: for the waveguide
        contact: for contacting the metal
        contact_width:
        contact_slab: function for the component contacting the slab
        contact_slab_top: Optional, defaults to contact_slab
        contact_slab_bot: Optional, defaults to contact_slab
        contact_slab_width: defaults to contact_width
        contact_spacing: spacing between contacts
        taper: optional taper
        kwargs: cross_section settings

    """
    c = Component()
    if taper:
        taper = taper() if callable(taper) else taper
        length -= 2 * taper.get_ports_xsize()

    wg = c << gf.c.straight(
        cross_section=cross_section,
        length=length,
        **kwargs,
    )

    contact_slab_width = contact_slab_width or contact_width
    contact_slab_spacing = contact_slab_spacing or contact_spacing

    if taper:
        t1 = c << taper
        t2 = c << taper
        t1.connect("o2", wg.ports["o1"])
        t2.connect("o2", wg.ports["o2"])
        c.add_port("o1", port=t1.ports["o1"])
        c.add_port("o2", port=t2.ports["o1"])

    else:
        c.add_ports(wg.get_ports_list())

    contact_length = length
    contact_top = c << contact(
        size=(contact_length, contact_width),
    )
    contact_bot = c << contact(
        size=(contact_length, contact_width),
    )

    contact_bot.x = wg.x
    contact_top.x = wg.x

    contact_top.ymin = +contact_spacing / 2
    contact_bot.ymax = -contact_spacing / 2

    contact_slab_top = contact_slab_top or contact_slab
    contact_slab_bot = contact_slab_bot or contact_slab
    slot_top = c << contact_slab_top(
        size=(contact_length, contact_slab_width),
    )
    slot_bot = c << contact_slab_bot(
        size=(contact_length, contact_slab_width),
    )

    slot_bot.x = wg.x
    slot_top.x = wg.x
    slot_top.ymin = +contact_slab_spacing / 2
    slot_bot.ymax = -contact_slab_spacing / 2

    c.add_ports(contact_bot.ports, prefix="bot_")
    c.add_ports(contact_top.ports, prefix="top_")
    return c


straight_pn_slot = gf.partial(straight_pin_slot, cross_section=pn)

if __name__ == "__main__":
    c = straight_pin_slot(contact_width=4, contact_slab_width=3, length=50)
    c.show()



from gdsfactory.component import Component
from gdsfactory.config import __version__


def grating_coupler(gc: Component) -> None:
    assert hasattr(
        gc.info, "polarization"
    ), f"{gc.name} does not have polarization attribute"
    assert gc.info.polarization in [
        "te",
        "tm",
    ], f"{gc.name} polarization  should be 'te' or 'tm'"
    assert hasattr(
        gc.info, "wavelength"
    ), f"{gc.name} wavelength does not have wavelength attribute"
    assert (
        0.5 < gc.info.wavelength < 5.0
    ), f"{gc.name} wavelength {gc.wavelength} should be in um"
    if "o1" not in gc.ports:
        print(f"grating_coupler {gc.name} should have a o1 port. It has {gc.ports}")
    if "o1" in gc.ports and gc.ports["o1"].orientation != 180:
        print(
            f"grating_coupler {gc.name} orientation = {gc.ports['o1'].orientation}"
            " should be 180 degrees"
        )


def version_equal_or_greater_than(version: str):
    """Raises error if version  is not high enough"""

    if __version__ < version:
        raise ValueError(
            f"gf.minimum required version = {version}\n"
            f"not compatible your current installed version {__version__}\n"
            "you can run:\n"
            "pip install gf.--upgrade\n"
            "to install a later version",
        )


if __name__ == "__main__":
    # equal_or_greater_than("3.0.0")
    # equal_or_greater_than("2.4.9")
    # equal_or_greater_than("2.4.8")
    version_equal_or_greater_than("2.5.0")



from typing import Callable, Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bezier import bezier
from gdsfactory.port import select_ports_optical


@gf.cell
def fanout2x2(
    component: Component,
    port_spacing: float = 20.0,
    bend_length: Optional[float] = None,
    npoints: int = 101,
    select_ports: Callable = select_ports_optical,
) -> Component:
    """returns component with port_spacing.

    Args:
        component: to package
        port_spacing: for the returned component
        bend_length: length of the bend (defaults to port_spacing)
        npoints: for sbend
        select_ports: function to select  optical_ports ports

    """

    c = gf.Component()

    component = component() if callable(component) else component
    component.component = component
    ref = c << component
    ref.movey(-ref.y)

    if bend_length is None:
        bend_length = port_spacing
    dx = bend_length

    y = port_spacing / 2.0

    p_w0 = ref.ports["o1"].midpoint
    p_w1 = ref.ports["o2"].midpoint
    p_e1 = ref.ports["o3"].midpoint
    p_e0 = ref.ports["o4"].midpoint
    y0 = p_e1[1]

    dy = y - y0

    control_points = [(0, 0), (dx / 2, 0), (dx / 2, dy), (dx, dy)]

    bezier_bend_t = bezier(
        control_points=control_points, npoints=npoints, start_angle=0, end_angle=0
    )

    b_tr = bezier_bend_t.ref(port_id="o1", position=p_e1)
    b_br = bezier_bend_t.ref(port_id="o1", position=p_e0, v_mirror=True)
    b_tl = bezier_bend_t.ref(port_id="o1", position=p_w1, h_mirror=True)
    b_bl = bezier_bend_t.ref(port_id="o1", position=p_w0, rotation=180)

    c.add([b_tr, b_br, b_tl, b_bl])

    c.add_port("o1", port=b_bl.ports["o2"])
    c.add_port("o2", port=b_tl.ports["o2"])
    c.add_port("o3", port=b_tr.ports["o2"])
    c.add_port("o4", port=b_br.ports["o2"])

    c.min_bend_radius = bezier_bend_t.info["min_bend_radius"]

    optical_ports = select_ports(ref.ports)
    for port_name in ref.ports.keys():
        if port_name not in optical_ports:
            c.add_port(port_name, port=ref.ports[port_name])
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    # c =gf.components.coupler(gap=1.0)
    c = gf.components.nxn(west=2, east=2)

    cc = fanout2x2(component=c)
    print(cc.ports["o3"].y - cc.ports["o4"].y)
    # print(cc.ports)
    cc.show()



from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.types import Layer


@cell
def triangle(
    x: float = 10,
    xtop: float = 0,
    y: float = 20,
    ybot: float = 0,
    layer: Layer = (1, 0),
) -> Component:
    r"""
    Args:
        x: base xsize
        xtop: top xsize
        y: ysize
        ybot: bottom ysize
        layer:

    .. code::

        xtop
           _
          | \
          |  \
          |   \
         y|    \
          |     \
          |      \
          |______|ybot
              x
    """
    c = Component()
    points = [[0, 0], [x, 0], [x, ybot], [xtop, y], [0, y]]
    c.add_polygon(points, layer=layer)
    return c


if __name__ == "__main__":
    cc = triangle(xtop=5, ybot=5)
    cc.show()



from pathlib import Path
from typing import Callable, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame

import gdsfactory as gf
from gdsfactory.simulation.write_sparameters_lumerical import (
    write_sparameters_lumerical,
)
from gdsfactory.types import Component, ComponentOrFactory


def plot_sparameters(
    component_or_df: Union[ComponentOrFactory, DataFrame],
    logscale: bool = True,
    keys: Optional[Tuple[str, ...]] = None,
    dirpath: Path = gf.CONFIG["sparameters"],
    write_sparameters_function: Callable = write_sparameters_lumerical,
    **sim_settings,
):
    """Plots Sparameters.

    Args:
        component_or_df: Component or Sparameters pandas DataFrame
        logscale: plots 20*log10(results)
        keys: list of keys to plot, plots all by default
        dirpath: where to store/read the simulations
        write_sparameters_function: custom function to write sparameters

    Keyword Args:
        sim_settings: simulation settings for the write_sparameters_function

    """

    df = component_or_df() if callable(component_or_df) else component_or_df
    if isinstance(df, Component):
        df = write_sparameters_function(component=df, dirpath=dirpath, **sim_settings)
    w = df["wavelength_nm"]

    keys = keys or [
        key for key in df.keys() if key.startswith("S") and key.endswith("m")
    ]

    for key in keys:
        if key in df:
            y = df[key]
            y = 20 * np.log10(y) if logscale else y
            plt.plot(w, y, label=key[:-1])
        else:
            raise ValueError(f"{key} not in {df.keys()}")
    plt.legend()
    plt.xlabel("wavelength (nm)")
    plt.ylabel("Transmission (dB)") if logscale else plt.ylabel("Transmission")


def plot_imbalance2x2(df: DataFrame, port1: str = "S13m", port2: str = "S14m") -> None:
    """Plots imbalance in % for 2x2 coupler"""
    y1 = df[port1].values
    y2 = df[port2].values
    imbalance = y1 / y2
    plt.plot(df.wavelength_nm, 100 * abs(imbalance))
    plt.xlabel("wavelength (nm)")
    plt.ylabel("imbalance (%)")
    plt.grid()


def plot_loss2x2(df: DataFrame, port1: str = "S13m", port2: str = "S14m") -> None:
    """Plots imbalance in % for 2x2 coupler"""
    y1 = df[port1].values
    y2 = df[port2].values
    plt.plot(df.wavelength_nm, abs(10 * np.log10(y1 ** 2 + y2 ** 2)))
    plt.xlabel("wavelength")
    plt.ylabel("excess loss (dB)")


plot_loss1x2 = gf.partial(plot_loss2x2, port1="S13m", port2="S12m")
plot_imbalance1x2 = gf.partial(plot_imbalance2x2, port1="S13m", port2="S12m")


if __name__ == "__main__":
    # plot_sparameters(df, logscale=True)
    # plot_sparameters(gf.components.coupler())
    plot_sparameters(gf.components.mmi1x2(), logscale=False)
    plt.show()



"""FIXME. Tests are failing for python3.7
"""


import gdsfactory as gf
from gdsfactory.component import hash_file


def test_hash_geometry() -> None:
    """Test geometric hash of the GDS points."""
    c1 = gf.components.straight(length=10)
    c2 = gf.components.straight(length=11)
    h1 = c1.hash_geometry()
    h2 = c2.hash_geometry()
    assert h1 != h2


def _test_hash_array_file():
    """Test hash of a component with an array of references."""
    gf.clear_cache()
    c = gf.Component("array")
    wg = gf.components.straight(length=3.2)
    c.add_array(wg)
    gdspath = c.write_gds()
    h = hash_file(gdspath)
    assert h == "71d476075cf081b4099c1eea1c8984a1", h


def _test_hash_file():
    """Test hash of the saved GDS file."""
    gf.clear_cache()
    c = gf.components.straight()
    c.add_label("hi")
    gdspath = c.write_gds()
    h = hash_file(gdspath)
    assert h == "f2228aed8141f447e601ce93a6219415", h


if __name__ == "__main__":
    # test_hash_geometry()
    _test_hash_file()
    _test_hash_array_file()



import sys
from pathlib import Path
from typing import Union

from gdsfactory.import_gds import import_gds


def remove_cell(component, cell_name) -> None:
    """Removes cell from a component"""
    all_cells = component.get_dependencies(recursive=True)
    all_cell_names = set([c.name for c in all_cells])
    if cell_name in all_cell_names:
        for c in all_cells:
            to_remove = []
            for e in c.references:
                to_remove += [e]
            for e in to_remove:
                # print("removing", e)
                c.remove(e)


def is_empty(c):
    # print(len(c.polygons), len(c.get_dependencies()))
    return len(c.polygons) == 0 and len(c.get_dependencies()) == 0


def remove_empty_cells(
    gds: Union[str, Path], recursive: bool = True, recurse_depth: int = 0
):
    """Returns the list cells to cells."""
    if isinstance(gds, (str, Path)):
        gds = import_gds(gds)

    cells = gds.get_dependencies(recursive=True)
    cells_to_remove = []
    for c in cells:
        if is_empty(c):
            cells_to_remove += [c]

    for c in cells_to_remove:
        print(c.name)
        remove_cell(gds, c.name)

    if recursive:
        while (len(cells_to_remove)) > 0:
            print(recurse_depth)
            sys.stdout.flush()
            cells_to_remove = remove_empty_cells(
                gds, recursive=True, recurse_depth=recurse_depth + 1
            )

    return cells_to_remove


def remove_empty_cells_from_gds_file(gdspath):
    component = import_gds(gdspath)
    remove_empty_cells(component)
    component.write_gds(gdspath[:-4] + "_cleaned.gds")


if __name__ == "__main__":
    remove_empty_cells_from_gds_file()



"""Read component GDS, YAML metadata and ports."""
from pathlib import Path
from typing import Union

from omegaconf import OmegaConf

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.import_gds import import_gds


def from_gds(gdspath: Union[str, Path], **kwargs) -> Component:
    """Returns Component with ports and metadata (YAML) info (if any).

    Args:
        gdspath: path of GDS file
        cellname: cell of the name to import (None) imports top cell
        flatten: if True returns flattened (no hierarchy)
        snap_to_grid_nm: snap to different nm grid (does not snap if False)

    """
    gdspath = Path(gdspath)
    metadata_filepath = gdspath.with_suffix(".yml")
    if not gdspath.exists():
        raise FileNotFoundError(f"No such file '{gdspath}'")
    component = import_gds(gdspath)

    if metadata_filepath.exists():
        metadata = OmegaConf.load(metadata_filepath)

        for port_name, port in metadata.ports.items():
            if port_name not in component.ports:
                component.add_port(
                    name=port_name,
                    midpoint=port.midpoint,
                    width=port.width,
                    orientation=port.orientation,
                    layer=port.layer,
                    port_type=port.port_type,
                )

        component.info = metadata.info
    return component


if __name__ == "__main__":
    gdspath = gf.CONFIG["gdsdir"] / "straight.gds"
    c = gf.read.from_gds(gdspath)
    c.show()
    print(c.to_yaml())

    # gdspath = gf.CONFIG["gdsdir"] / "straight.gds"
    # c = gf.c.straight()
    # c.write_gds_with_metadata(gdspath)

    # gdspath = gf.CONFIG["gdsdir"] / "straight.gds"
    # metadata_filepath = gdspath.with_suffix(".yml")
    # m = OmegaConf.load(metadata_filepath)
    # c2 = import_gds(gdspath)
    # c2.show()
